/* dhcp4_lease.c
 *
 *  Representation of ISC dhcp lease options for libdhcp
 *
 *  Copyright(C) Jason Vas Dias<jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#ifndef _DHCP4_LEASE_H
#define _DHCP4_LEASE_H

#include  <sys/types.h>
#include  <netinet/in.h>
#include  <stdint.h>

/*
 *  DHCP Option Type Codes:
 */
typedef
enum dhcp_option_type__e
{   /* These are the same codes as used by dhcpd / dhclient,
     * and which are written by dhclient with '-x' option to 
     * new option information variables named x._universe_.x:
     */
    /* "atomic" types: */
    DHC_T_IP_ADDRESS        = 'I', /*    I - IP address */
    DHC_T_HEX_STRING        = 'X', /*    X - either an ASCII string or binary data.   On output, the string is */
                                   /*        scanned to see if it's printable ASCII and, if so, output as a */
                                   /*        quoted string.   If not, it's output as colon-seperated hex.   On */
                                   /*        input, the option can be specified either as a quoted string or as */
                                   /*        a colon-seperated hex list. */
    DHC_T_DOMAIN_NAME       = 'd', /*    d - Domain name (i.e., FOO or FOO.BAR). */
    DHC_T_INT32             = 'l', /*    l - 32-bit signed integer */
    DHC_T_UINT32            = 'L', /*    L - 32-bit unsigned integer */
    DHC_T_INT16             = 's', /*    s - 16-bit signed integer */
    DHC_T_UINT16            = 'S', /*    S - 16-bit unsigned integer */
    DHC_T_CHAR              = 'b', /*    b - 8-bit signed integer */
    DHC_T_UCHAR             = 'B', /*    B - 8-bit unsigned integer */
    DHC_T_TEXT              = 't', /*    t - ASCII text */
    DHC_T_BOOL              = 'f', /*    f - flag (true or false) */
    DHC_T_IMPLICIT_BOOL     = 'F', /*    F - implicit flag - the presence of the option indicates that the */

    /* compound types: */
    DHC_T_ARRAY             = 'A', /*    A - array of whatever precedes (e.g., IA means array of IP addresses) */
    DHC_T_LIST              = 'a', /*    a - array of the preceding character (e.g., IIa means two or more IP */
                                   /*        addresses) */
    /* there are NO enumeration type dhcp options defined 
     * and no means of specifying user defined enumeration types -
     * the 'parameters.state' option we deal with as an integer or text
     */
    DHC_T_ENUMERATION       = 'N', /*    N - enumeration.   N is followed by a text string containing */
                                   /*        the name of the set of enumeration values to parse or emit, */
                                   /*        followed by a '.'.   The width of the data is specified in the */
                                   /*        named enumeration.   Named enumerations are tracked in parse.c. */
    /* these are types internal to DHCP and can  safely be ignored .
     * ( they are not emitted by dhclient in environment -
     *   except "vendor-option-space", which we deal with by use of
     *   '._universe_.' information.
     * )
     */
    DHC_T_OPTION_SPACE      = 'U', /*    U - name of an option space (universe) */

                                   /*        flag is true. */
    DHC_T_OPTIONAL          = 'o', /*    o - the preceding value is optional. */
    DHC_T_ENCAPSULATION     = 'E', /*    E - encapsulation, string or colon-seperated hex list (the latter */
                                   /*        two for parsing).   E is followed by a text string containing */
                                   /*        the name of the option space to encapsulate, followed by a '.'. */
                                   /*        If the E is immediately followed by '.', the applicable vendor */
                                   /*        option space is used if one is defined. */
    DHC_T_ENCAPSULATION_NO_DIRECTIVE= 'e',
                                   /*    e - If an encapsulation directive is not the first thing in the string, */
                                   /*        the option scanner requires an efficient way to find the encapsulation. */
                                   /*        This is done by placing a 'e' at the beginning of the option.   The */
                                   /*        'e' has no other purpose, and is not required if 'E' is the first */
                                   /*        thing in the option. */
    DHC_T_NONE               = '\0'
}   DHCPv4_Option_Type;

enum dhcpv4_option_structure_type_e
{
    DHCO_SCALAR,
    DHCO_ARRAY,
    DHCO_STRUCT
}   DHCPv4_Option_Structure_Type;

#define DHCPV4_MAX_OPTION_LENGTH  1236

typedef struct dhcp4_option_s
{   /* Generic option structure */
    struct dhcp4_lease_s *lease;      /* the lease this option is associated with */
    char *                name;       /* dhcp option name, defined in dhcp/common/tables.c */
    char *                format;     /* dhcp format string of DHCPv4_Option_Type(s) */
    char *                universe;   /* universe name */
    uint8_t               unicode;    /* universe code */
    uint8_t               code;       /* 0 - 255 */
    uint8_t               form;       /* DHCPv4_Option_Structure_Type */
    uint8_t               size;       /* If struct / array, size of element */
    uint8_t               udefined;   /* 1 if user defined option */
    uint8_t               redefined;  /* 1 if user redefined option*/
    uint16_t              n_members;  /* > 1 if structure */ 
    uint16_t              n_elements; /*> 1 if array */   
    uint16_t              length;     /* size of value in bytes */
    uint8_t               **members;  /* points to array of pointers to members at end of value */
    uint8_t               value[1];   /* option value(s) c-structure ; must be aligned for integer */
} DHCPv4_option;

/* structured options (options with more than one member)
 * are laid out as C structures in 'opt->value', with a
 * list of pointers to each member following the C structure.
 * opt->length will be :
 * sizeof(structure) + (n_members * sizeof(void*)).
 * ie. sizeof(structure) 
 *     == opt->length -(n_members * sizeof(void*)).
 *
 * Array options have exactly opt->n_members elements of opt->size,
 * and are packed as a c array.
 *
 * For scalars, n_members == 1.
 */

typedef  void (*DHCPv4_option_handler)(DHCPv4_option *, void* );

typedef struct dhcp4_lease_s
{
    struct in_addr address;           /* client IP address being leased    */ 
    struct in_addr requested_address; /* address we requested, (if any)    */
    struct in_addr server_address;    /* boot server IP address (if any)   */
    char *filename;                   /* boot filename                     */
    char *server_name;                /* boot server name                  */
    uint8_t is_static;                /* 1: was from config, not pool      */
    uint8_t is_bootp;                 /* 1: is a bootp lease               */
    uint16_t if_index;                /* lease on interface with this index*/
    char *if_name;                    /* lease on interface with this name */
    time_t  requested,                /* time that lease was requested     */
            expiry, renewal, rebind;  /* times in seconds since epoch      */
    void *options;                    /* glibc tsearch(3) btree of options */
    void *options_by_name;            /* glibc tsearch(3) btree of options */
    DHCPv4_option_handler handler;    /* handler for this lease            */
    void *handler_arg;
} DHCPv4_lease;

struct client_state;                  /* include "dhcpd.h" to get this     */

extern DHCPv4_lease *dhcpv4_lease( struct client_state * );
    /* creates a DHCPv4_lease from an ISC DHCP 'struct client_state'.
     */
extern void dhcpv4_lease_free( DHCPv4_lease * );

extern void dhcpv4_process_options ( DHCPv4_lease *, DHCPv4_option_handler, void *); 

extern DHCPv4_option *dhcpv4_get_option_by_code( DHCPv4_lease *, uint8_t universe, uint8_t code);

extern DHCPv4_option *dhcpv4_get_option_by_name( DHCPv4_lease *, char *); 


/* DHCP Code Spaces:
 *
 * Option Codes have unique values in one of these "DHCP_Universe" 
 * code spaces:
 */

#define DHCP_OPTION_MAX   256
/*  DHCP can define no more than 256 options, 0-255, in a "universe" */

#define DHCP_UNIVERSE_MAX 10
/*  DHCP currently supports no more than 10 universes (a hardcoded constant in tables.c).
 *  It defines 4 of its own and supports up to 5 user defined option spaces.
 */

/*  "Universe" codes defined by ISC DHCP: */
typedef 
enum dhcp_u_e
{
    DHC_DHCP_Universe=0,           /* Proper DHCP Options as defined in RFC 1533 */
    DHC_RAI_Universe=1,            /* Relay Agent Information sub-option encapsulation */
    DHC_NWIP_Universe=2,           /* NetWare/IP sub-option encapsulation */
    DHC_FQDN_Universe=3,           /* Fully Qualified Domain Name sub-option encapsulation */
    DHC_DEFINED_UNIVERSE_N=4
} DHCP_Universe;

/*  DHCP Universe  Option Codes: 
 */
typedef
enum dhcp_oc_e
{
    DHCO_PAD				=	0, /* Unused */
    DHCO_SUBNET_MASK			=	1,
    DHCO_TIME_OFFSET			=	2,
    DHCO_ROUTERS			=	3,
    DHCO_TIME_SERVERS			=	4,
    DHCO_NAME_SERVERS			=	5,
    DHCO_DOMAIN_NAME_SERVERS		=	6,
    DHCO_LOG_SERVERS			=	7,
    DHCO_COOKIE_SERVERS			=	8,
    DHCO_LPR_SERVERS			=	9,
    DHCO_IMPRESS_SERVERS		=	10,
    DHCO_RESOURCE_LOCATION_SERVERS	=	11,
    DHCO_HOST_NAME			=	12,
    DHCO_BOOT_SIZE			=	13,
    DHCO_MERIT_DUMP			=	14,
    DHCO_DOMAIN_NAME			=	15,
    DHCO_SWAP_SERVER			=	16,
    DHCO_ROOT_PATH			=	17,
    DHCO_EXTENSIONS_PATH		=	18,
    DHCO_IP_FORWARDING			=	19,
    DHCO_NON_LOCAL_SOURCE_ROUTING	=	20,
    DHCO_POLICY_FILTER			=	21,
    DHCO_MAX_DGRAM_REASSEMBLY		=	22,
    DHCO_DEFAULT_IP_TTL			=	23,
    DHCO_PATH_MTU_AGING_TIMEOUT		=	24,
    DHCO_PATH_MTU_PLATEAU_TABLE		=	25,
    DHCO_INTERFACE_MTU			=	26,
    DHCO_ALL_SUBNETS_LOCAL		=	27,
    DHCO_BROADCAST_ADDRESS		=	28,
    DHCO_PERFORM_MASK_DISCOVERY		=	29,
    DHCO_MASK_SUPPLIER			=	30,
    DHCO_ROUTER_DISCOVERY		=	31,
    DHCO_ROUTER_SOLICITATION_ADDRESS	=	32,
    DHCO_STATIC_ROUTES			=	33,
    DHCO_TRAILER_ENCAPSULATION		=	34,
    DHCO_ARP_CACHE_TIMEOUT		=	35,
    DHCO_IEEE802_3_ENCAPSULATION	=	36,
    DHCO_DEFAULT_TCP_TTL		=	37,
    DHCO_TCP_KEEPALIVE_INTERVAL		=	38,
    DHCO_TCP_KEEPALIVE_GARBAGE		=	39,
    DHCO_NIS_DOMAIN			=	40,
    DHCO_NIS_SERVERS			=	41,
    DHCO_NTP_SERVERS			=	42,
    DHCO_VENDOR_ENCAPSULATED_OPTIONS	=	43,
    DHCO_NETBIOS_NAME_SERVERS		=	44,
    DHCO_NETBIOS_DD_SERVER		=	45,
    DHCO_NETBIOS_NODE_TYPE		=	46,
    DHCO_NETBIOS_SCOPE			=	47,
    DHCO_FONT_SERVERS			=	48,
    DHCO_X_DISPLAY_MANAGER		=	49,
    DHCO_DHCP_REQUESTED_ADDRESS		=	50,
    DHCO_DHCP_LEASE_TIME		=	51,
    DHCO_DHCP_OPTION_OVERLOAD		=	52,
    DHCO_DHCP_MESSAGE_TYPE		=	53,
    DHCO_DHCP_SERVER_IDENTIFIER		=	54,
    DHCO_DHCP_PARAMETER_REQUEST_LIST	=	55,
    DHCO_DHCP_MESSAGE			=	56,
    DHCO_DHCP_MAX_MESSAGE_SIZE		=	57,
    DHCO_DHCP_RENEWAL_TIME		=	58,
    DHCO_DHCP_REBINDING_TIME		=	59,
    DHCO_VENDOR_CLASS_IDENTIFIER	=	60,
    DHCO_DHCP_CLIENT_IDENTIFIER		=	61,
    DHCO_NWIP_DOMAIN_NAME		=	62,
    DHCO_NWIP_SUBOPTIONS		=	63,
    DHCO_NISPLUS_DOMAIN                 =       64,
    DHCO_NISPLUS_SERVERS                =       65,
    DHCO_TFTP_SERVER_NAME               =       66,
    DHCO_BOOTFILE_NAME                  =       67,
    DHCO_MOBILE_IP_HOME_AGENT           =       68,
    DHCO_SMTP_SERVER                    =       69,
    DHCO_POP_SERVER                     =       70,
    DHCO_NNTP_SERVER                    =       71,
    DHCO_WWW_SERVER                     =       72,
    DHCO_FINGER_SERVER                  =       73,
    DHCO_IRC_SERVER                     =       74,
    DHCO_STREETTALK_SERVER              =       75,
    DHCO_STREETTALK_DIRECTORY_ASSISTANCE_SERVER=76,    
    DHCO_USER_CLASS			=	77,
    DHCO_SLP_DIRECTORY_AGENT            =       78,
    DHCO_SLP_SERVICE_SCOPE              =       79,
    DHCO_FQDN				=	81,
    DHCO_DHCP_AGENT_OPTIONS		=	82,
    DHCO_NDS_SERVERS                    =       85,
    DHCO_NDS_TREE_NAME                  =       86,
    DHCO_NDS_CONTEXT                    =       87,
    DHCO_UAP_SERVERS                    =       98,     
    DHCO_SUBNET_SELECTION		=	118,
    DHCO_AUTHENTICATE			=	210,
    DHCO_END                            =       210,
    DHCO_N                              =       211
}   DHCP_Option_Code;

/* DHCP Relay Agent Information Sub-Options:
 */
typedef
enum dho_rai_c_e
{
    DHCO_RAI_PAD                         =       0,
    DHCO_RAI_CIRCUIT_ID	                 =       1,
    DHCO_RAI_REMOTE_ID	                 =       2,
    DHCO_RAI_AGENT_ID	                 =       3,
    DHCO_RAI_END                         =       3,
    DHCO_RAI_N                           =       4
} DHCP_RAI_Options;

/* DHCP NetWare/IP Sub-Options:
 */
typedef
enum dho_nwip_c_e
{
    DHCO_NWIP_PAD                        =       0,
    DHCO_NWIP_ILL1                       =       1,
    DHCO_NWIP_ILL2                       =       2,
    DHCO_NWIP_ILL3                       =       3,
    DHCO_NWIP_ILL4                       =       4,
    DHCO_NWIP_NSQ_BROADCAST              =       5,
    DHCO_NWIP_PREFERRED_DSS              =       6,
    DHCO_NWIP_NEAREST_NWIP_SERVER        =       7,
    DHCO_NWIP_AUTORETRIES                =       8,
    DHCO_NWIP_AUTORETRY_SECS             =       9,
    DHCO_NWIP_1_1                        =      10,
    DHCO_NWIP_PRIMARY_DSS                =      11,
    DHCO_NWIP_END                        =      11,
    DHCO_NWIP_N                          =      12
} DHCP_NWIP_Options;

/* DHCP Fully Qualified Domain Name Sub-Options:
 */
typedef
enum dho_fqdn_c_e
{
    DHCO_FQDN_PAD                        =       0,
    DHCO_FQDN_NO_CLIENT_UPDATE		 =       1,
    DHCO_FQDN_SERVER_UPDATE		 =       2,
    DHCO_FQDN_ENCODED			 =       3,
    DHCO_FQDN_RCODE1			 =       4,
    DHCO_FQDN_RCODE2			 =       5,
    DHCO_FQDN_HOSTNAME			 =       6,
    DHCO_FQDN_DOMAINNAME		 =       7,
    DHCO_FQDN_FQDN			 =       8,
    DHCO_FQDN_END                        =       8,
    DHCO_FQDN_N                          =	 9
} DHCP_FQDN_Options;

/* NOTE: programs using DHCP must be able to handle (or ignore)
 * user defined options in new universes with user defined formats,
 * and handle existing defined options that have been redefined by
 * the user (ISC dhcp allows this!).
 */
#endif
