/* dhcp6_nic.c
 *
 * Network Interface Configurator for the DHCPv6 IPv6 DHCP client library.
 *
 *  Copyright(C) Jason Vas Dias <jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <malloc.h>
#include <stdio.h>

#include <dhcp6_nic.h>
extern int dhcpv6_client(LIBDHCP_Control *dhc_ctl, int argc, char **argv, char **envp);

/*
typedef
struct dhcp6_nic_s
{
    NLH_t          nh;
    NIC_t          nic;
    IPaddr_list_t  address_list;
    IPaddr_list_t  dns_list;
    char          *search_list;
    DHCPv6_lease  *lease;
} DHCPv6_nic;
*/

static void
dhc6_log( LIBDHCP_Control *ctl, int priority, char *fmt, ...)
{
    va_list va;
    va_start(va, fmt);
    ctl->eh(ctl, priority, fmt, va);
    va_end(va);
}

extern char **environ;

DHCPv6_nic *do_dhcpv6
(   LIBDHCP_Control *ctl, NLH_t nh, char *if_name   )
{
    DHCPv6_nic *nic6 = calloc(1, sizeof(DHCPv6_nic));
    ctl->arg = nic6;
    nic6->nh = nh;
    nic6->nic = nic_by_name(nh, if_name);
    nic6->name= strdup(if_name);
    dhc6_log(ctl, LOG_INFO,"DHCPv6: NIC: %p %s %p", nic6, if_name, nic6->nic);
    STAILQ_INIT(&(nic6->address_list));
    STAILQ_INIT(&(nic6->route_list));
    STAILQ_INIT(&(nic6->dns_list));
    ctl->arg = nic6;
    libdhcp_call_client
    (  ctl, dhcpv6_client, "dhcp6c",
       "-Df", if_name, 0 
    );    
    return ctl->arg;
}

NIC_Res_t dhcpv6_nic( DHCPv6_nic *nic6 )
{
    return 
	nic_configure
	(   nic6->nh,
	    nic6->nic,
	    &(nic6->address_list),
	    &(nic6->route_list), /* no routes configured by DHCPv6 - for that, use radvd! */
	    &(nic6->dns_list),
	    nic6->search_list,
	    0 /* no host name set - use host(1) to query PTR */
        );
}

extern
int dhcp6_nic_callback
(   LIBDHCP_Control *ctl,
    DHCP_State state, 
    void *arg
)
{
    DHCPv6_nic *nic = ctl->arg;
    char buf[64];

    dhc6_log
    (   ctl, LOG_INFO, "DHCPv6 %s - state: %s",
	nic->name,
	libdhcp_state_string(state,buf)
    );

    switch (state)
    {
    case DHC_TIMEDOUT:
	dhc6_log
        (   ctl, LOG_ERR, "DHCPv6 %s - TIMED OUT - failure.", nic->name
        );
	ctl -> finished = 1;
	break;

    case DHC6_BOUND:
    case DHC6_REBIND:
	{   /* record new lease */
	    dhc6_log
	    (   ctl, LOG_INFO, "DHCPv6 %s - BOUND", nic->name );
	    nic->lease = dhcpv6_lease( arg );
	    ctl -> finished = 1;    
	}
	break;

    case DHC6_RELEASE:
	/* remove existing lease */
	ctl -> finished = 1;
	break;

    default:
	dhc6_log
        (   ctl, LOG_DEBUG, "DHCPv6 %s - unhandled state.", nic->name );
    }
    return 0;
}

int dhcp6_process_lease( LIBDHCP_Control *ctl, DHCPv6_nic *nic )
{
    if( ( nic == 0L ) || (nic->lease == 0L) ) return(0);
    DHCPv6_lease *lease = nic->lease;
    DHCPv6_lease_address *addr=0 ;
    char buf[128];
    IPaddr_t ip;
    IPaddr_list_node_t *n;
    int addresses = 0;

    STAILQ_FOREACH( addr, &(lease->addresses), link)
    {
	dhc6_log
	(   ctl, LOG_INFO, "DHCPv6 %s - lease address: %s/%d",  nic->name,
	    inet_ntop(AF_INET6,&(addr->address),buf,sizeof(buf)),
	    addr->prefix_length
	);
	struct sockaddr_in6 sa6 ;
	memset(&sa6,'\0', sizeof(struct sockaddr_in6));
	memcpy(&(sa6.sin6_addr), &(addr->address), sizeof(struct in6_addr));
	sa6.sin6_family = AF_INET6;
	ip = nic_sa_addr(nic->nh, (struct sockaddr*)&(sa6), sizeof(struct in6_addr));
	nic_addr_set_prefix(ip, addr->prefix_length );
	n = calloc(1, sizeof(IPaddr_list_node_t));
	n->addr = ip;
	STAILQ_INSERT_TAIL(&(nic->address_list), n, link);
	addresses++;
    }
    if ( !STAILQ_EMPTY( &(lease->dns) ) )
    {
	IPv6_address *dns;
	STAILQ_FOREACH(dns, &(lease->dns), link)
	{
	    dhc6_log
	    (   ctl, LOG_INFO, "DHCPv6 %s - dns server address: %s", nic->name,
		inet_ntop(AF_INET6,&(dns->address),buf,sizeof(buf))
	    );
	    struct sockaddr_in6 sa6 ;
	    memset(&sa6,'\0', sizeof(struct sockaddr_in6));
	    memcpy(&(sa6.sin6_addr), &(dns->address), sizeof(struct in6_addr));
	    sa6.sin6_family = AF_INET6;
	    ip = nic_sa_addr(nic->nh, (struct sockaddr*)&(sa6), sizeof(struct in6_addr));
	    n = calloc(1, sizeof(IPaddr_list_node_t));
	    n->addr = ip;		    
	    STAILQ_INSERT_TAIL(&(nic->dns_list), n, link);
	}
    }
    if ( lease->search_list )
    {
	dhc6_log
	(   ctl, LOG_INFO, "DHCPv6 %s - dns search_path: %s", 	nic->name,
	    lease->search_list
	);
	nic->search_list = strdup( lease->search_list );		
    }
    return addresses;
}
