/* dhcp_nic.c
 *
 * Network Interface Configurator for BOTH the ISC DHCP IPv4 client library
 * and the DHCPv6 IPv6 client.
 *
 *  Copyright(C) Jason Vas Dias <jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#include <sys/types.h>
#include <unistd.h>
#include <dhcp_nic.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <pthread.h>

static
int dhc_log( LIBDHCP_Control *ctl, int priority, char *fmt, ... )
{
    if( (ctl==0) || (ctl->eh == 0) )
	return 0;
    va_list va;
    va_start(va,fmt);
    int r = ctl->eh(ctl, priority, fmt, va);
    va_end(va);
    return r;
}

static
int dhc_log_noctl( LIBDHCP_Error_Handler eh, int log_level, int priority, char *fmt, ... )
{        
    LIBDHCP_Control ctl;
    if( eh == 0 )
	return 0;
    ctl.log_level = log_level;
    va_list va;
    va_start(va,fmt);
    int r = eh(&ctl, priority, fmt, va);
    va_end(va);
    return r;
}

static int ifup( LIBDHCP_Control *control, NIC_t nic )
{
    /* bring interface UP */
    uint32_t iff = nic_get_flags(nic);
    dhc_log
    (   control, LOG_INFO,
	"DHCP %s  - bringing link UP - %x %d",
	nic_get_name(nic),
	iff, iff & (IFF_UP | IFF_RUNNING) 
    );
    if ( ( iff & (IFF_UP | IFF_RUNNING) ) != (IFF_UP | IFF_RUNNING) ) 
    {
	if ( nic_set_flags(nic, iff | IFF_UP | IFF_RUNNING ) != NIC_OK )
	{
	    dhc_log
	    (   control, LOG_ERR,
		"DHCP %s - bringing link UP failed.",
		nic_get_name(nic)
	    );
	    return 1;
	}
	sleep(2); /* give kernel time to assign IPv6 link-local address ! */
    }
    return 0;
}

#ifdef DHCP_USE_THREADS
struct dhcp6_thx
{
    void *ctl, *nh, *ethX;
};

static 
void *dhcpv6_thread( void *arg )
{
    struct dhcp6_thx *thx = (struct dhcp6_thx*)arg;

    void *dhc6_nic  = do_dhcpv6(thx->ctl, thx->nh, thx->ethX);
    fprintf(stderr,"\n\n--- DHCPv6 nic: %p\n", dhc6_nic);    
    pthread_exit(dhc6_nic);
    return dhc6_nic;
}

static void *dhcpv4_thread( void *dhc4ctl )
{
    void *dhc4_nic= do_dhcpv4( dhc4ctl );

    fprintf(stderr,"\n\n--- DHCPv4 nic: %p\n", dhc4_nic);
    pthread_exit(dhc4_nic);
    return dhc4_nic;    
}
#endif

extern 
DHCP_nic *do_dhcp
( 
    DHCP_Preference preference,
    char                  *ethX,
    LIBDHCP_Capability     dhc_cap, 
    time_t                 timeout,    
    LIBDHCP_Error_Handler  error_handler,
    uint8_t                log_level,
    ...
)
{    
    va_list dhclient_va;
    va_start(dhclient_va, log_level);

    DHCP_nic *nic = calloc(1,sizeof(DHCP_nic));
    if( nic == 0L )
    {
	dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: out of memory");
	return 0L;
    }

    nic->preference = preference;

    NLH_t nh = 
	nic_open(0);

    nic->nh = nh;

    if ( nh == 0L )
    {
	dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: nic_open() failed");
	free(nic);
	return 0L;
    }
    
    NIC_t eth = nic_by_name(nh, ethX);

    if( eth == 0L )
    {
	dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: net_get_by_name(%s) failed", ethX);
	free(nic);
	return 0L;
    }

    LIBDHCP_Control *dhc6ctl=0;
    DHCPv4_control  *dhc4ctl=0;

    if ( (preference & DHCPv4_DISABLE) == 0 )
    {
	dhc4ctl =
	    dhcpv4_control_va
	    (nh, ethX, dhc_cap, 10, libdhcp_stderr_logger, log_level, dhclient_va);
	if ( dhc4ctl == 0L )
	{
	    dhc_log_noctl(error_handler, log_level, LOG_FATAL,"dhcp_nic: dhcpv4_control failed");
	    free(nic);
	    nic_close(&nh);
	    return 0;
	}

	nic->dhc4ctl = dhc4ctl;

    }
    if ( (preference & DHCPv6_DISABLE) == 0 )
    {
	dhc6ctl = 
	    libdhcp_control_new	      
	    (   dhcp6_nic_callback,
		0,  /* NO "capabilities"! */
		10, /* timeout */
		0,  /* let do_dhcpv6 worry about arg */
		libdhcp_stderr_logger, LOG_INFO
	    );

	if ( dhc6ctl == 0L )
	{
	    dhc_log_noctl(error_handler, log_level, LOG_FATAL,"dhcp_nic: libdhcp_control_new failed");
	    free(nic);
	    nic_close(&nh);
	    return 0;
	}

	nic->dhc6ctl = dhc6ctl;

    }
    if ( ( dhc6ctl == 0L ) && (dhc4ctl == 0) )
    {
	dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: both DISABLE_DHCPv6 and DISABLE_DHCv4 set.");
	free(nic);
	nic_close(&nh);
	return 0;
    }

    if( error_handler )
    {
	nic_set_va_logger(nh, (NIC_VA_Error_Handler_t)error_handler, dhc6ctl);
	nic_set_loglevel( nh, log_level );
    }


    if( ifup(dhc6ctl, eth) != 0 )
    {
	dhc_log_noctl(error_handler, log_level, LOG_FATAL, "dhcp_nic: ifup(%s) failed", ethX);
	free(nic);
	return 0L;
    }

    DHCPv4_nic *nic4=0;
    DHCPv6_nic *nic6=0;

#ifdef DHCP_USE_THREADS
    if( (preference & (DHCPv6_DISABLE | DHCPv4_DISABLE)) == 0 )
    {	
	pthread_t  dhcp4_thread = 0, dhcp6_thread = 0;
	struct dhcp6_thx thx = { nic->dhc6ctl, nh, ethX };
	pthread_create( &dhcp4_thread, NULL, dhcpv4_thread, dhc4ctl );
	pthread_create( &dhcp6_thread, NULL, dhcpv6_thread, &thx );
	pthread_join( dhcp4_thread, (void**)&nic4 );
	pthread_join( dhcp6_thread, (void**)&nic6 );
	fprintf(stderr,"nic4: %p\nnic6: %p\n", nic4, nic6);
    }else
#endif
    {
    if ( (preference & DHCPv4_DISABLE) == 0 )
	nic4 = do_dhcpv4(nic->dhc4ctl);

    if ( (preference & DHCPv6_DISABLE) == 0 )
	nic6 = do_dhcpv6(nic->dhc6ctl, nh, ethX);

    }

    if ( nic4 )
	if ( ! dhcp4_process_lease( dhc4ctl ) )
	    dhc_log_noctl(error_handler, log_level, LOG_ERR,
			  "DHCP: empty DHCPv4 lease."
		         );
    if ( nic6 )
	if ( ! dhcp6_process_lease( dhc6ctl, nic6 ) )
	    dhc_log_noctl(error_handler, log_level, LOG_ERR,
			  "DHCP: empty DHCPv6 lease."
		         );
    
    nic->dhcpv4 = (DHCP_config*)nic4;
    nic->dhcpv6 = (DHCP_config*)nic6;
    return nic;
}

NIC_Res_t
dhcp_nic_configure( DHCP_nic *nic )
{
    if (nic == 0)
	return NIC_FAIL;

    LIBDHCP_Control *ctl = nic->dhc6ctl;
    NIC_Res_t res = NIC_FAIL;    
    IPaddr_list_t      sdns, *dns=&sdns;
    IPaddr_list_node_t *n=0;
    char *search_list = 0;

    if ( ( nic -> dhcpv6 ) 
       &&( nic -> dhcpv4 )
       )
    { /* We have both IPv6 and IPv4 configurations -
       * The only potential conflicts are with resolv.conf -
       * first merge IPv4 and IPv6 resolver configurations
       * according to preference :
       */ 	

	STAILQ_INIT(dns);

	if( (nic->preference & DHCPv6_DISABLE_RESOLVER)
	  || STAILQ_EMPTY( &(nic->dhcpv6->dns_list) )
	  )
	    dns = (nic->preference & DHCPv4_DISABLE_RESOLVER)
	          ? 0
                  : &(nic->dhcpv4->dns_list);
	else
	if( (nic->preference & DHCPv4_DISABLE_RESOLVER) 
          || STAILQ_EMPTY( &(nic->dhcpv4->dns_list) )
	  )
	    dns = (nic->preference & DHCPv6_DISABLE_RESOLVER)
		 ? 0
		 : &(nic->dhcpv6->dns_list);
	else
	if( nic->preference & IPv6_PREFERENCE )
	{
	    STAILQ_FOREACH(n, &(nic->dhcpv6->dns_list), link)
	    {
		STAILQ_INSERT_TAIL(dns, n, link);
	    }

	    STAILQ_FOREACH(n, &(nic->dhcpv4->dns_list), link)
	    {
		STAILQ_INSERT_TAIL(dns, n, link);
	    }       

	    if ( nic->dhcpv6->search_list == 0 )
		search_list = (nic->preference & DHCPv4_DISABLE_RESOLVER) 
		              ? 0 
		              : nic->dhcpv4->search_list;
	    else
            if ( nic->dhcpv4->search_list == 0 )
		search_list = (nic->preference & DHCPv6_DISABLE_RESOLVER) 
		              ? 0
		              : nic->dhcpv6->search_list;		              
	    else
	    {
		search_list =
		    malloc
		    (   strlen(nic->dhcpv6->search_list)
		      + strlen(nic->dhcpv4->search_list)
		      + 2
		    );
		sprintf(search_list,"%s %s", 
			nic->dhcpv6->search_list, 
			nic->dhcpv4->search_list
		       );
	    }
	   	    
	}else /* (preference & IPv6_PREFERENCE) == 0 */
	{
	    if( (nic->preference & DHCPv4_DISABLE_RESOLVER) == 0 )
	    {
		STAILQ_FOREACH(n, &(nic->dhcpv4->dns_list), link)
		{
		    STAILQ_INSERT_TAIL(dns, n, link);
		}
	    }
	    if( (nic->preference & DHCPv6_DISABLE_RESOLVER) == 0 )
	    {
		STAILQ_FOREACH(n, &(nic->dhcpv6->dns_list), link)
		{
		    STAILQ_INSERT_TAIL(dns, n, link);
		}
	    }
	    if ( nic->dhcpv4->search_list == 0 )
	    {
		search_list = (nic->preference & DHCPv6_DISABLE_RESOLVER) 
		              ? 0
		              : nic->dhcpv6->search_list;
	    }else
            if ( nic->dhcpv6->search_list == 0 )
	    {
		search_list = (nic->preference & DHCPv4_DISABLE_RESOLVER) 
		              ? 0
		              : nic->dhcpv4->search_list;
	    }else
	    {
		search_list =
		    malloc
		    (   strlen(nic->dhcpv4->search_list)
		      + strlen(nic->dhcpv6->search_list)
		      + 2
		    );
		sprintf(&(search_list[0]),"%s %s", 
			nic->dhcpv4->search_list, 
			nic->dhcpv6->search_list
		       );
	    }
	}

	if( nic->preference & IPv6_PREFERENCE )
	{
	    /* DHCPv6 */
	    res=nic_configure
		(
		    nic->nh,
		    nic->dhcpv6->nic,
		    (nic->preference & DHCPv6_DISABLE_ADDRESSES) ? 0 : &(nic->dhcpv6->address_list),
		    0,   /* never any dhcpv6 routes ! (use radvd for that.) */
		    dns,
		    search_list,
		    0    /* dhvpv6 never sets host name */
		);

	    if( res != NIC_OK )
	    {
		dhc_log
		    ( ctl, LOG_ERR, 
		      "DHCPv6 interface configuration failed."
		    );
		return res;
	    }else
	    {
		dhc_log
		    ( ctl, LOG_INFO, 
		      "DHCPv6 interface configuration succeeded."
		    );
	    }

	    /* DHCPv4 */
	    res=nic_configure
		(
		    nic->nh,
		    nic->dhcpv4->nic,
		    (nic->preference & DHCPv4_DISABLE_ADDRESSES) ? 0 : &(nic->dhcpv4->address_list),
		    (nic->preference & DHCPv4_DISABLE_ROUTES)    ? 0 : &(nic->dhcpv4->route_list),
		    dns,
		    search_list,
		    (nic->preference & DHCPv4_DISABLE_HOSTNAME_SET)?0: nic->dhcpv4->host_name
		);

	    if( res != NIC_OK )
	    {
		dhc_log
		    ( ctl, LOG_ERR, 
		      "DHCPv4 interface configuration failed."
		    );
		return res;
	    }else
	    {
		dhc_log
		    ( ctl, LOG_INFO, 
		      "DHCPv4 interface configuration succeeded."
		    );
	    }

	}else	
	{
	    /* DHCPv4 */
	    res=nic_configure
		(
		    nic->nh,
		    nic->dhcpv4->nic,
		    (nic->preference & DHCPv4_DISABLE_ADDRESSES) ? 0 : &(nic->dhcpv4->address_list),
		    (nic->preference & DHCPv4_DISABLE_ROUTES)    ? 0 : &(nic->dhcpv4->route_list),
		    dns,
		    search_list,
		    (nic->preference & DHCPv4_DISABLE_HOSTNAME_SET)?0: nic->dhcpv4->host_name
		);

	    if( res != NIC_OK )
	    {
		dhc_log
		    ( ctl, LOG_ERR, 
		      "DHCPv4 interface configuration failed."
		    );
		return res;

	    }else
	    {
		dhc_log
		    ( ctl, LOG_INFO, 
		      "DHCPv4 interface configuration succeeded."
		    );
	    }

	    /* DHCPv6 */
	    res=nic_configure
		(
		    nic->nh,
		    nic->dhcpv6->nic,
		    (nic->preference & DHCPv6_DISABLE_ADDRESSES) ? 0 : &(nic->dhcpv6->address_list),
		    0,   /* never any dhcpv6 routes ! (use radvd for that.) */
		    dns,
		    search_list,
		    0
		);

	    if( res != NIC_OK )
	    {
		dhc_log
		    ( ctl, LOG_ERR, 
		      "DHCPv6 interface configuration failed."
		    );
		return res;
	    }else
	    {
		dhc_log
		    ( ctl, LOG_INFO, 
		      "DHCPv6 interface configuration succeeded."
		    );
	    }

	}
    } else            /* only one configuration */
    if ( nic->dhcpv4 )
    {
	/* DHCPv4 */
	res=nic_configure
	    (
		nic->nh,
		nic->dhcpv4->nic,
		(nic->preference & DHCPv4_DISABLE_ADDRESSES) ? 0 : &(nic->dhcpv4->address_list),
		(nic->preference & DHCPv4_DISABLE_ROUTES)    ? 0 : &(nic->dhcpv4->route_list),
		(nic->preference & DHCPv4_DISABLE_RESOLVER)   ? 0 : &(nic->dhcpv4->dns_list),
	        (nic->preference & DHCPv4_DISABLE_RESOLVER)   ? 0 : nic->dhcpv4->search_list,
		(nic->preference & DHCPv4_DISABLE_HOSTNAME_SET)?0 : nic->dhcpv4->host_name
	    );

	if( res != NIC_OK )
	{
	    dhc_log
		(   ctl, LOG_ERR, 
		    "DHCPv4 interface configuration failed."
		);
	    return res;
	}else
	{
	    dhc_log
	        (   ctl, LOG_INFO, 
		    "DHCPv4 interface configuration succeeded."
		);
	}

    } else
    if ( nic->dhcpv6 )
    {
	/* DHCPv6 */
	res=nic_configure
	    (
		nic->nh,
		nic->dhcpv6->nic,
		(nic->preference & DHCPv6_DISABLE_ADDRESSES) ? 0 : &(nic->dhcpv6->address_list),
		0,   /* never any dhcpv6 routes ! (use radvd for that.) */
		(nic->preference & DHCPv6_DISABLE_RESOLVER)  ? 0 : &(nic->dhcpv6->dns_list),
	        (nic->preference & DHCPv6_DISABLE_RESOLVER)  ? 0 : nic->dhcpv6->search_list,
		0
	    );

	if( res != NIC_OK )
	{
	    dhc_log
		(   ctl, LOG_ERR, 
		    "DHCPv6 interface configuration failed."
		);
	    return res;
	}else
	{
	    dhc_log
	        (   ctl, LOG_INFO, 
		    "DHCPv6 interface configuration succeeded."
		);
	}	
    }
    return res;
}
