/* nic.c
 *
 *  Network Interface Configurator
 *  
 *
 *  Copyright(C) Jason Vas Dias <jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#include <sys/types.h>
#include <sys/unistd.h>
#include <nic.h>
#include <net/if_arp.h>
#include <arpa/inet.h>
#include <sys/syslog.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <search.h>
extern   void tdestroy (void *root, void (*free_node)(void *nodep));

#include <libdhcp.h>

/* libnl Netlink API headers: */

#include <netlink/netlink.h>
#include <netlink/msg.h>
#include <netlink/attr.h>
#include <netlink/utils.h>
#include <netlink/addr.h>
#include <netlink/route/rtnl.h>
#include <netlink/route/link.h>
#include <netlink/route/addr.h>
#include <netlink/route/route.h>
extern int nl_get_errno(void);

struct nlh_s
{
    struct nl_handle *nl;   /* netlink handle */
    struct nl_cache  *lc;   /* link cache     */
    struct nl_cache  *ac;   /* address cache  */ 
    struct nl_cache  *rc;   /* route cache    */ 
    int               ll;   /* log level      */    
    NIC_Error_Handler_t    eh;       /* error handler / logger */
    NIC_VA_Error_Handler_t va_eh;    /* underlying va_list logger (libdhcp) */
    void                  *va_eh_arg;/* va_list logger 1st argument */
};

static void flush_caches(NLH_t nh);

static int init_caches(NLH_t nh);

#define nic_alloc( sz )  malloc( sz )
#define nic_calloc(n,sz) calloc( n, sz )
#define nic_fre( p )     free(p)

NLH_t nic_open(NIC_Error_Handler_t eh)
{
    NLH_t nh = nic_calloc(1, sizeof(struct nlh_s));
    
    if ( nh == 0L )
	return 0L;

    nh->nl = nl_handle_alloc();
    if ( nh->nl == 0L )
    {
	if( eh )
	    eh ( 0, NIC_FATAL, "cannot allocate NIC library handle" );
	goto nic_open_fail;
    }
    
    if ( nl_connect(nh->nl, NETLINK_ROUTE)  < 0 ) 
    {
	 if( eh )
	     eh (0, NIC_FATAL, "cannot connect netlink socket" );
	 goto nic_open_fail;
    }

    if( init_caches( nh ) != 0 )
	goto nic_open_fail;
    
    if( eh )
    {
	nh->ll = NIC_ERR;
	nh->eh = eh;
    }
    return( nh );

nic_open_fail:
    nic_close(&nh);
    return( 0L );
}

void nic_close(NLH_t *nh)
{
    if( nh == 0 )
	return;

    flush_caches(*nh);

    if( (*nh)->nl )
	nl_close( (*nh)->nl );       

    nic_fre(*nh);
    nh = 0;
}

static 
int init_caches(NLH_t nh)
{
    if( (nh == 0L) || (nh->nl == 0L) )
	return(-1);
    
    flush_caches(nh);

    if ( (nh->lc = rtnl_link_alloc_cache( nh->nl )) == 0L )
    {
	if( nh->eh )
	    nh->eh (0, NIC_FATAL, "cannot allocate link cache - %d : %s\n",
		    nl_get_errno(), nl_geterror()
	           );
	 goto init_caches_fail;
    }

    if ( (nh->ac = rtnl_addr_alloc_cache( nh->nl )) == 0L )
    {
	if( nh->eh )
	    nh->eh (0, NIC_FATAL, "cannot allocate address cache - %d : %s\n",
		    nl_get_errno(), nl_geterror()
		   );
	 goto init_caches_fail;
    }

    if ( (nh->rc = rtnl_route_alloc_cache( nh->nl )) == 0L )
    {
	if( nh->eh )
	    nh->eh (0, NIC_FATAL, "cannot allocate route cache - %d : %s\n",
		    nl_get_errno(), nl_geterror()
		   );
	 goto init_caches_fail;
    }    
    return ( 0 );
init_caches_fail:
    return ( -1);
}

static void flush_caches(NLH_t nh)
{
    if( (nh == 0L) || (nh->nl == 0L) )
	return;
    if( nh->lc )
    {
	nl_cache_destroy_and_free(nh->lc);
	nh->lc = 0;
    }
    if( nh->ac )
    {
	nl_cache_destroy_and_free(nh->ac);
	nh->ac = 0L;
    }
    if( nh->rc )
    {
	nl_cache_destroy_and_free(nh->rc);
	nh->rc = 0L;
    }
}

/* default error handlers: */

void nic_sys_logger( NLH_t nh, NIC_Error_Level_t el, char *fmt, ... )
{
    if ( (nh == 0L) || ( el > nh->ll ) )
	return;
    va_list va;
    
    va_start(va, fmt);
    
    if( nh->va_eh )
	nh->va_eh ( nh->va_eh_arg, el, fmt, va );
    else 
	vsyslog( el, fmt, va );

    va_end(va);

    if ( el == LOG_FATAL )
	nic_close(&nh);
}

void nic_stderr_logger( NLH_t nh, NIC_Error_Level_t el, char *fmt, ... )
{
    if ( (nh == 0L) || ( el > nh->ll ) )
	return;

    va_list va;
    
    va_start(va, fmt);
    if( nh->va_eh )
	nh->va_eh ( nh->va_eh_arg, el, fmt, va );
    else 
    {
	vfprintf(stderr, fmt, va);
	fprintf(stderr,"\n");
	fflush(stderr);
    }
    va_end(va);
}

NIC_Res_t nic_set_loglevel( NLH_t nh, NIC_Error_Level_t ll )
{
    if ( nh == 0 )
	return NIC_FAIL;

    nh -> ll = ll;
    
    return NIC_OK;
}

extern
void nic_set_logger(NLH_t nh, NIC_Error_Handler_t eh)
{
    nh -> eh = eh;
}

extern
void nic_set_va_logger(NLH_t nh, NIC_VA_Error_Handler_t va_eh, void *va_eh_arg)
{
    nh -> va_eh = va_eh;
    nh -> va_eh = va_eh;
}

struct nic_s
{
    NLH_t             nh;
    struct rtnl_link *link;    
};

NIC_t nic_by_name(NLH_t nh, char *if_name)
{    
    if( (if_name == 0L) || ( *if_name == '\0' ) )
    {
	if( nh -> eh )
	    nh -> eh ( nh, NIC_FATAL, "nic_by_name: invalid nil argument" );
	return( 0L );	
    }

    NIC_t nic = nic_calloc(1, sizeof(struct nic_s));
    
    if( nic == 0L )
    {
	if( nh -> eh )
	    nh -> eh ( nh, NIC_FATAL, "nic_by_name: out of memory" );
	return( 0L );
    }

    if( ( nic -> link = rtnl_link_get_by_name( nh->lc, if_name ) ) == 0L )
    {
	if( nh -> eh )
	    nh -> eh ( nh, NIC_WARN, "nic_by_name: no such link named '%s' - %d: %s", if_name,
		       nl_get_errno(), nl_geterror()
		     );
	nic_fre(nic);
	return( 0L );
    }

    nic -> nh = nh;
    return nic;
}

NIC_t nic_by_index(NLH_t nh, uint16_t if_index)
{    
    NIC_t nic = nic_calloc(1, sizeof(struct nic_s));
    
    if( nic == 0L )
    {
	if( nh -> eh )
	    nh -> eh ( nh, NIC_FATAL, "nic_by_index: out of memory" );
	return( 0L );
    }

    if( ( nic -> link = rtnl_link_get( nh->lc, if_index ) ) == 0L )
    {
	if( nh -> eh )
	    nh -> eh ( nh, NIC_WARN, "nic_by_index: no such link with index %d - %d: %s", if_index,
		       nl_get_errno(), nl_geterror()
		     );
	return( 0L );
    }

    nic -> nh = nh;
    return nic;
}

struct nic_cbarg
{
    void (*cb)(NLH_t, NIC_t,  void *);
    void  *arg;
    NLH_t nh;
};

static void link_cache_cb(struct nl_object *nlo,  void *arg)
{
    struct nic_cbarg *cba = arg;
    struct rtnl_link *link = (struct rtnl_link*)nlo;   
    struct nic_s snic = { cba->nh, link };
    
    cba->cb(cba->nh, &snic, cba->arg);
}

void nic_foreach(NLH_t nh, NIC_handler_t handler,  void *arg)
{
    struct nic_cbarg cba = { handler, arg, nh };
    nl_cache_foreach(nh->lc,link_cache_cb,&cba);
}

char* nic_get_name( NIC_t nic )
{
    return rtnl_link_get_name(nic->link);
}

uint16_t nic_get_index( NIC_t nic )
{
    return rtnl_link_get_ifindex(nic->link);
}

struct nic_ip_address_s 
{
    NLH_t nh;
    struct rtnl_addr *a;
    struct nl_addr *addr;
    uint8_t sa_len;
    uint8_t sa[ (_SS_SIZE - _SS_PADSIZE) & 0xff ];
    uint8_t sa_data[3]; /*  _SS_PADSIZE  */
};
#define IPaddr_sockaddr(p) ((struct sockaddr*)&(p->sa[0]))
#define IPaddr_socklen(p)  (p->sa_len + (_SS_SIZE - _SS_PADSIZE))
typedef struct nic_ip_address_s IP_t;
#define IPADDR_FIXED_LEN   (sizeof(IP_t) - 3)

typedef
enum { LLADDR, LLBRD } LLADDR_TYPE;

static IPaddr_t nic_get_link_layer_addr( NIC_t nic, LLADDR_TYPE ll_addr_type )
{
    if( nic == 0L )
	return 0L;

    struct nl_addr *nl_addr = 
	ll_addr_type == LLADDR 
	? rtnl_link_get_addr( nic->link ) 
	: rtnl_link_get_broadcast( nic->link );
    uint8_t *ba, sa_len, sa_family;
    IP_t *r;

    if ( nl_addr == 0L )
	goto nic_get_link_addr_fail;
    
    sa_family = nl_addr_get_family( nl_addr );

    sa_len = nl_addr_get_len( nl_addr );
    
    if( sa_len <= 0 )
	goto nic_get_link_addr_fail;
    
    if ( (ba = nl_addr_get_binary_addr( nl_addr )) == 0L )
	goto nic_get_link_addr_fail;
    
    r = nic_calloc(1, sizeof(IP_t) + sa_len);

    r->nh = nic->nh;
    r->sa_len = sa_len;
    IPaddr_sockaddr(r)->sa_family = sa_family;
    memcpy( &(((struct sockaddr*)&(r->sa[0]))->sa_data[0]), ba, sa_len);
/*    nl_addr_put(nl_addr);*/
    return r;

nic_get_link_addr_fail:
    
    if( nic->nh->eh )
	nic->nh->eh (nic->nh, NIC_ERR, "nic_get_link_addr failed: no link address: %d %s\n", 
		     nl_get_errno(), nl_geterror()
	            );
    return 0;
}

IPaddr_t nic_get_link_addr(NIC_t nic)
{
    return nic_get_link_layer_addr( nic, LLADDR );
}

IPaddr_t nic_get_link_broadcast(NIC_t nic)
{
    return nic_get_link_layer_addr( nic, LLBRD );
}

uint32_t nic_get_flags( NIC_t nic )
{
    if(nic == 0L)
	return NIC_FAIL;
    return rtnl_link_get_flags( nic->link );
}

NIC_Res_t nic_set_flags( NIC_t nic, uint32_t flags )
{
    int r;

    if( (nic == 0L) || (nic->link == 0) )
	return NIC_FAIL;

    uint8_t ix = rtnl_link_get_ifindex(nic->link);
    
    struct rtnl_link *link_new = rtnl_link_alloc();
    if ( link_new == 0L )
    {
	if( nic->nh->eh )
	    nic->nh->eh (nic->nh, NIC_ERR, "out of memory");
	return NIC_FAIL;
    }
    rtnl_link_set_flags(link_new,flags);
    if ( (r=rtnl_link_change(nic->nh->nl, nic->link, link_new,0)) < 0 )
    {
	if( nic->nh->eh )
	    nic->nh->eh (nic->nh, NIC_ERR, "cannot set link flags: netlink error: %d %d %s\n", 
			 r, nl_get_errno(), nl_geterror()
		        );
	return NIC_FAIL;
    }
    rtnl_link_put(nic->link);
    rtnl_link_put(link_new);
    nic->link =  rtnl_link_get(nic->nh->lc,ix);
    return NIC_OK;
}

uint32_t nic_get_mtu( NIC_t nic )
{
    if(nic == 0L)
	return NIC_FAIL;
    return rtnl_link_get_mtu( nic->link );
}

NIC_Res_t nic_set_mtu( NIC_t nic, uint32_t mtu )
{
    int r;

    if( nic == 0L )
	return NIC_FAIL;
    
    struct rtnl_link *link_new = rtnl_link_alloc();
    if ( link_new == 0L )
    {
	if( nic->nh->eh )
	    nic->nh->eh (nic->nh, NIC_ERR, "out of memory");
	return NIC_FAIL;
    }
    rtnl_link_set_mtu(link_new,mtu);
    if ( (r=rtnl_link_change(nic->nh->nl, nic->link, link_new, 0)) < 0 )
    {
	if( nic->nh->eh )
	    nic->nh->eh (nic->nh, NIC_ERR, "cannot set link mtu: netlink error: %d %d %s\n", 
			 r, nl_get_errno(), nl_geterror()
		        );
	return NIC_FAIL;
    }
    rtnl_link_put( nic->link );
    nic->link = link_new;
    return NIC_OK;
}

struct rtnl_link *nic_rtnl_link_get( NIC_t nic)
{
    return nic->link;
}

void nic_rtnl_link_put( struct rtnl_link* link)
{
    rtnl_link_put(link);
}

void nic_free( NIC_t nic )
{
    if( nic->link )
	rtnl_link_put(nic->link);
    nic_fre(nic);
}


static IPaddr_t ipaddr( NLH_t nh, struct rtnl_addr *addr, IP_t *ip, size_t len )
{
    struct sockaddr *sa=(struct sockaddr*)&(ip->sa[0]);
    struct nl_addr *nla = rtnl_addr_get_local( addr );
    uint8_t sa_len,*ba;
    if( (nla == 0L) || (ip == 0L) )
	return 0;
    ba = nl_addr_get_binary_addr(nla);
    if( ba == 0L )
	return 0;
    sa_len = nl_addr_get_len(nla);
    if( sa_len == 0 )
	return 0;
    if( (((unsigned long)ip) + len) < (((unsigned long)(&(ip->sa_data[0])))+sa_len))
	return 0;    
    ip->nh = nh;
    ip->a=addr;
    ip->addr = nla;
    ip->sa_len = sa_len;
    sa->sa_family = nl_addr_get_family(nla);
    switch( sa->sa_family )
    {
    case AF_INET:	
	memcpy(&(((struct sockaddr_in*)sa)->sin_addr), ba, sa_len);  
	break;
    case AF_INET6:
	memcpy(&(((struct sockaddr_in6*)sa)->sin6_addr), ba, sa_len);  
	break;
    default:
	memcpy(&(((struct sockaddr*)sa)->sa_data[0]), ba, sa_len);  
	break;
    }
    return ip;
}

struct sockaddr *
nic_addr_sa( IPaddr_t ip, socklen_t *sz)
{
    if( sz == 0L )
	return 0L;
    *sz = IPaddr_socklen(ip);
    return IPaddr_sockaddr(ip);
}

IPaddr_t nic_sa_addr(NLH_t nh, struct sockaddr *sa, socklen_t sz )
{
    if( (sa == 0L) || (sa->sa_family == AF_UNSPEC) || (sz <= 0) )
	return 0;
    size_t tlen = IPADDR_FIXED_LEN + sz;
    IP_t *ip = nic_calloc(1,tlen);
    if( ip == 0L )
	return(0L);
    ip->nh = nh;
    ip->a = rtnl_addr_alloc();
    if( ip->a == 0L )
    {
	nic_fre(ip);
	return(0L);
    }
    struct nl_addr *nla = nl_addr_alloc(_SS_SIZE);
    if( nla == 0L )
    {
	rtnl_addr_free(ip->a);
	nic_fre(ip);
	return(0L);
    }
    struct sockaddr *ipsa = IPaddr_sockaddr(ip);
    ipsa -> sa_family = sa->sa_family;
    nl_addr_set_family(nla, sa->sa_family);
    switch( sa->sa_family )
    {
    case AF_INET:
	memcpy(&(((struct sockaddr_in*)ipsa)->sin_addr), 
	       &(((struct sockaddr_in*)sa)->sin_addr),
	       sz
	      );
	nl_addr_set_binary_addr
	    (nla,
	     &(((struct sockaddr_in*)ipsa)->sin_addr),
	     sz
	    );
	ip->sa_len = sz;
	break;
    case AF_INET6:
	memcpy(&(((struct sockaddr_in6*)ipsa)->sin6_addr), 
	       &(((struct sockaddr_in6*)sa)->sin6_addr),
	       sz
	      );
	nl_addr_set_binary_addr
	    (nla,
	     &(((struct sockaddr_in6*)ipsa)->sin6_addr),
	     sz
	    );
	ip->sa_len = sz;
	break;
    default:
	memcpy(&(ipsa->sa_data[0]),&(sa->sa_data[0]), sz);
	nl_addr_set_binary_addr(nla, &(ipsa->sa_data[0]), sz);
	ip->sa_len = sz;
	break;
    }
    rtnl_addr_set_family(ip->a,sa->sa_family);
    rtnl_addr_set_local(ip->a,nla);
    rtnl_addr_set_scope(ip->a,RT_SCOPE_UNIVERSE);
    return ip;
}

extern
IPaddr_t nic_ipv4_addr( NLH_t nh, struct in_addr ip4 )
{
    struct sockaddr_in sin;
    memset(&(sin),'\0',sizeof(struct sockaddr_in));
    sin.sin_family = AF_INET;
    sin.sin_addr = ip4;
    return nic_sa_addr(nh, (struct sockaddr*)&sin, sizeof(sin.sin_addr));    
}

IPaddr_t nic_ipv4_broadcast_addr( IPaddr_t ip )
{
    if( (ip == 0) || (ip->sa_len == 0) )
	return 0L;
    struct sockaddr *sa = IPaddr_sockaddr(ip);
    if( sa->sa_family != AF_INET )
	return 0L;
    uint8_t prefix = nic_addr_get_prefix( ip );
    if( prefix == 0 )
	return 0L;
    struct sockaddr_in sin, *isin=(struct sockaddr_in*)sa;
    memset(&sin, '\0', sizeof(struct sockaddr_in));
    uint32_t ip4 = ntohl(isin->sin_addr.s_addr);
    sin.sin_addr.s_addr =
    htonl( ( ip4 & (((1 << prefix)-1) << (32 - prefix)))
	                         |(((1<< (32 - prefix))-1))
	 );
    sin.sin_family = AF_INET;
    return nic_sa_addr(ip->nh, (struct sockaddr*)&sin, sizeof(struct in_addr));
}

NIC_Res_t nic_addr_get_broadcast( IPaddr_t ip, struct sockaddr *ba, size_t sz  )
{
    if( ( ip == 0 ) || (ip->a == 0) )
	return 0;

    struct nl_addr *nla = rtnl_addr_get_broadcast ( ip->a );
    if ( nla == 0L )
	return NIC_FAIL;
    
    uint8_t *da = nl_addr_get_binary_addr( nla );

    if ( da == 0L )
	return NIC_FAIL;
    
    uint8_t len = nl_addr_get_len( nla );
    
    if ( ( len == 0 ) || (len > sz) )
	return NIC_FAIL;

    uint8_t fa = nl_addr_get_family( nla );

    if ( fa == AF_UNSPEC )
	return NIC_FAIL;

    memset(ba,'\0', sizeof(struct sockaddr));
    switch(fa)
    {
    case AF_INET:
	memcpy(&(((struct sockaddr_in*)ba)->sin_addr), da, len);
	((struct sockaddr_in*)ba)->sin_family = AF_INET;
	break;
    case AF_INET6:
	memcpy(&(((struct sockaddr_in6*)ba)->sin6_addr), da, len);
	((struct sockaddr_in6*)ba)->sin6_family = AF_INET6;
	break;
    default:
	memcpy(&(((struct sockaddr*)ba)->sa_data[0]), da, len);
	((struct sockaddr*)ba)->sa_family = fa;
	break;
    }
    return NIC_OK;
}

void nic_addr_set_broadcast( IPaddr_t ip, IPaddr_t ipb )
{
    if((ip==0)||(ipb==0))
	return;
    struct sockaddr *sa = IPaddr_sockaddr(ipb);
    struct nl_addr *nla = nl_addr_alloc(_SS_SIZE);    
    nl_addr_set_family(nla, sa->sa_family );
    switch( sa->sa_family )
    {
    case AF_INET:	
	nl_addr_set_binary_addr( nla, &(((struct sockaddr_in*)sa)->sin_addr), ipb->sa_len);
	break;
    case AF_INET6:
	nl_addr_set_binary_addr( nla, &(((struct sockaddr_in6*)sa)->sin6_addr), ipb->sa_len);
	break;
    default:
	nl_addr_set_binary_addr( nla, &(sa->sa_data[0]), ipb->sa_len);
    }
    rtnl_addr_set_broadcast( ip->a, nla );
}

void nic_addr_set_scope(IPaddr_t ip, int8_t scope)
{
    if((ip==0L)||(ip->a==0L))
	return;
    rtnl_addr_set_scope(ip->a, scope);
}

int8_t nic_addr_get_scope(IPaddr_t ip)
{
    if((ip==0L)||(ip->a==0L))
	return 0;
    return rtnl_addr_get_scope(ip->a);
}

extern
IPaddr_t nic_ipv6_addr( NLH_t nh, struct in6_addr ip6 )
{
    struct sockaddr_in6 sin6;
    memset(&(sin6),'\0',sizeof(struct sockaddr_in6));
    sin6.sin6_family = AF_INET;
    sin6.sin6_addr = ip6;
    return nic_sa_addr(nh, (struct sockaddr*)&sin6, sizeof(sin6.sin6_addr));    
}

uint8_t nic_addr_get_family( IPaddr_t ip )
{
    if( ip->a )
	return rtnl_addr_get_family(ip->a);
    else 
	return IPaddr_sockaddr(ip)->sa_family;
}

uint8_t nic_addr_get_prefix( IPaddr_t ip )
{
    if( ip->a )
	return rtnl_addr_get_prefixlen(ip->a);
    else
	return 0;
}

void nic_addr_set_prefix( IPaddr_t ip, uint8_t prefix )
{
    if( ip->a )
	rtnl_addr_set_prefixlen(ip->a, prefix);

}

IPaddr_t nic_addr_from_text(NLH_t nh, char *as )
{
    struct sockaddr_storage ss;
    struct sockaddr *sa=(struct sockaddr*)&ss;
    uint8_t prefix=0;    
    char addr_buf[256], *p;
    socklen_t sz =0;
    strcpy(addr_buf, as);
    as = addr_buf;
    p = strchr(as,'/');
    if( p )
	*p='\0';
    memset(sa, '\0', sizeof(struct sockaddr_storage));
    if( inet_pton(AF_INET6, as, &(((struct sockaddr_in6*)sa)->sin6_addr)) )
    {
	sa->sa_family = AF_INET6;
	sz = sizeof(struct in6_addr);
    }else
    if( inet_pton(AF_INET, as, &(((struct sockaddr_in*)sa)->sin_addr)) )
    {
	sa->sa_family = AF_INET;
	sz = sizeof(struct in_addr);
    }
    if ( sa->sa_family )
    {
	IPaddr_t ip = nic_sa_addr(nh, sa, sz);
	if(p)
	{
	    if(ip)
	    {
		int pfx =0;
		if( sscanf(p+1,"%3d",&pfx) )
		{
		    prefix = (uint8_t) pfx;
		    nic_addr_set_prefix(ip,prefix);
		}
		*p='/';
	    }
	}
	return ip;
    }else
	return 0L;
}

static void hex_dump( uint8_t *src, uint32_t src_len, uint8_t *dst, uint32_t dst_len, char sep)
{
    while( src_len-- && (dst_len-=3) )
	dst+=sprintf((char*)dst,"%.2x%c",*(src++),sep); 
    *(--dst)='\0';
}

char *nic_addr_to_text( IPaddr_t ip, char *s, size_t sz)
{
    char sep='.';
    if( ( ip == 0L ) || (ip->sa_len == 0))	
	return 0L;
	
    struct sockaddr *sa = IPaddr_sockaddr(ip);

    switch( sa->sa_family )
    {
    case AF_INET:
	if( ( s == 0L ) && ( (s = nic_alloc( sz = 32 )) == 0) )
	    return 0L;
	inet_ntop(AF_INET, &(((struct sockaddr_in*)sa)->sin_addr), s, sz);
	break;

    case AF_INET6:
	if( ( s == 0L ) && ( (s = nic_alloc( sz = 64 )) == 0))
	    return (char*)0;
	inet_ntop(AF_INET6,&(((struct sockaddr_in6*)sa)->sin6_addr), s, sz);
	break;

    case AF_LLC:
	sep=':';
    default:
	if( ( s == 0L ) && ( (s = nic_alloc( sz = ((ip->sa_len * 3)+1))) == 0))
	    return (char*)0;
	hex_dump((uint8_t*)&(sa->sa_data[0]), ip->sa_len, (uint8_t*)s, sz, sep);	
    }
    return s;
}

struct ipaddr_cbarg
{
    void (*cb)(NLH_t nh, IPaddr_t,  void *);
    void  *arg;
    uint16_t ix;
    NLH_t nh;
};

static void addr_cache_cb(struct nl_object *nlo,  void *arg)
{
    struct ipaddr_cbarg *cba = arg;
    struct rtnl_addr *addr = (struct rtnl_addr*) nlo;       
    if( cba->ix )
	if( rtnl_addr_get_ifindex(addr) != cba->ix )
	    return;
    uint8_t buf[ _SS_SIZE + sizeof(IP_t) ];
    IPaddr_t saddr = ipaddr(cba->nh, addr, (IPaddr_t)&(buf[0]),  _SS_SIZE + sizeof(IP_t));
    if( saddr == 0 )
	return;   
    cba->cb(cba->nh, saddr, cba->arg);
}

void nic_addr_foreach( NLH_t nh, IPaddr_Handler_t handler, void* arg)
{
    struct ipaddr_cbarg cba = { handler, arg, 0, nh };
    nl_cache_foreach(nh->ac,addr_cache_cb,&cba);
}

void nic_addr_nic_foreach( NLH_t nh, NIC_t nic, IPaddr_Handler_t handler, void* arg)
{
    struct ipaddr_cbarg cba = { handler, arg, rtnl_link_get_ifindex(nic->link), nh };
    nl_cache_foreach(nh->ac,addr_cache_cb,&cba);
}

NIC_Res_t nic_add_address( NIC_t nic, IPaddr_t ip )
{
    if(( nic == 0L ) || (ip == 0L) || (ip->a == 0L) )
	return NIC_FAIL;

    uint16_t ix = rtnl_link_get_ifindex( nic->link );

    rtnl_addr_set_ifindex( ip->a, ix );

    if( rtnl_addr_add( nic->nh->nl, ip->a, NLM_F_REPLACE ) < 0 )
    {
	if( nic->nh->eh )
	    nic->nh->eh ( nic->nh, NIC_ERR, "failed to add address - %d : %s",
			  nl_get_errno(), nl_geterror()
		        );
	return NIC_FAIL;
    }
    return NIC_OK;
}

NIC_Res_t nic_remove_address( NIC_t nic, IPaddr_t ip )
{
    if(( nic == 0L ) || (ip == 0L) || (ip->a == 0L) )
	return NIC_FAIL;

    uint16_t ix = rtnl_link_get_ifindex( nic->link );

    rtnl_addr_set_ifindex( ip->a, ix );

    if( rtnl_addr_delete( nic->nh->nl, ip->a, 0 ) < 0 )
    {
	if( nic->nh->eh )
	    nic->nh->eh ( nic->nh, NIC_ERR, "failed to remove address - %d : %s",
			  nl_get_errno(), nl_geterror()
		        );
	return NIC_FAIL;
    }
    return NIC_OK;
}

IPaddr_list_t *nic_address_list_new( IPaddr_t ip1, ... )
{      
    va_list va;
    va_start(va, ip1);
    
    IPaddr_list_t *head = nic_calloc(1,sizeof(IPaddr_list_t));
    IPaddr_list_node_t *n;
    IP_t *ip;

    if( head == 0 )
    {
	va_end(va);
	if( ip1->nh->eh )
	    ip1->nh->eh ( ip1->nh, NIC_ERR, "out of memory" );
	return 0;
    }
    ip=ip1;
    STAILQ_INIT(head);
    do
    {
	n = nic_calloc(1,sizeof(IPaddr_list_node_t));
	if ( n == 0L )
	{
	    va_end(va);
	    nic_fre(head);
	    if( ip->nh->eh )
		ip->nh->eh ( ip->nh, NIC_ERR, "out of memory" );
	    return 0L;
	}
	n->addr = ip;
	STAILQ_INSERT_TAIL( head, n, link );
    }while( (ip = va_arg(va, IP_t*)) != 0) ;
    va_end(va);
    return(head);
}

void nic_address_list_free( IPaddr_list_t *head )
{
    IPaddr_list_node_t *n, *a;
    
    if ( !STAILQ_EMPTY(head) )
    {
	n = STAILQ_FIRST( head );
	do
	{
	    a = STAILQ_NEXT(n, link);
	    nic_addr_free(n->addr);
	    nic_fre(n);
	    n = a;
	}while (n);
    }
}

NIC_Res_t nic_add_addresses( NIC_t nic, IPaddr_list_t *head )
{
    IPaddr_list_node_t *n;
    NIC_Res_t r;

    STAILQ_FOREACH(n,head,link)
    {
	if((r = nic_add_address(nic, n->addr)) == NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

NIC_Res_t nic_remove_addresses( NIC_t nic, IPaddr_list_t *head )
{
    IPaddr_list_node_t *n;
    NIC_Res_t r;

    STAILQ_FOREACH(n,head,link)
    {
	if((r = nic_remove_address(nic, n->addr)) == NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

struct rtnl_addr *nic_rtnl_addr_get(IPaddr_t ip)
{
    if( ip->a )
	return ip->a;
    return 0L;
}

struct nl_addr *nic_nl_addr_get(IPaddr_t ip)
{
    if( ip->addr )
	return ip->addr;
    return 0L;
}

void nic_rtnl_addr_put(struct rtnl_addr *ra)
{
    if( ra != 0 )
	rtnl_addr_put(ra);
}

void nic_nl_addr_put(struct nl_addr *na)
{
    if( na != 0 )
	nl_addr_put(na);
}

void nic_addr_free( IPaddr_t ip )
{
    if( ip == 0 )
	return;
    if( ip->a != 0 )
	rtnl_addr_put( ip->a );
    if( ip->addr != 0 )
	nl_addr_put(ip->addr);
    nic_fre(ip);
}


struct nic_route_s
{
    NLH_t nh;
    struct rtnl_route  *rt;    
    int32_t table;
    int16_t oif;
    int16_t iif;
    uint8_t dst_len;
    uint8_t src_len;
    uint8_t family;
    uint8_t scope;
    uint32_t flags;    
    uint32_t priority;
    struct sockaddr dst_sa; 
    uint8_t pad1[ sizeof(struct sockaddr_in6) - sizeof(struct sockaddr) ];
    struct sockaddr gw_sa; 
    uint8_t pad2[ sizeof(struct sockaddr_in6) - sizeof(struct sockaddr) ];
    struct sockaddr src_sa; 
    uint8_t pad3[ sizeof(struct sockaddr_in6) - sizeof(struct sockaddr) ];
};

struct route_cbarg
{
    NLH_t nh;
    IProute_handler_t cb;
    void *arg;
};

static void route_cache_cb(struct nl_object *nlo,  void *arg)
{
    struct route_cbarg  *cba = arg;
    struct rtnl_route *route = (struct rtnl_route*) nlo;       
    struct nic_route_s rt;
    memset(&rt, '\0', sizeof(rt));
    rt.rt = route;
    rt.nh = cba->nh;
    cba->cb(cba->nh, &rt, cba->arg);
}

void nic_route_foreach( NLH_t nh, IProute_handler_t handler, void* arg)
{
    struct route_cbarg cba = { nh, handler, arg };
    nl_cache_foreach(nh->rc,route_cache_cb,&cba);
}

int32_t nic_route_get_table( IProute_t rt)
{
    if( rt == 0L ) return 0;
    
    if( rt->rt == 0L )
	return rt->table;

    return (rt->table = rtnl_route_get_table(rt->rt));
}

char *nic_route_get_table_name( int32_t table, char *buf, int sz )
{
    return rtnl_route_table2str( table, buf, sz );
}

int32_t nic_route_get_table_number( char *table )
{
    return rtnl_route_str2table( table );
}

uint8_t nic_route_get_family( IProute_t rt)
{
    if( rt == 0L ) return 0;
    
    if( rt->rt == 0L )
	return rt->family;

    return (rt->family = rtnl_route_get_family(rt->rt));
}

int16_t nic_route_get_oif( IProute_t rt )
{
    if( rt == 0L ) return 0xffff;
    
    if( rt->rt == 0L )
	return rt->oif;

    return (rt->oif = rtnl_route_get_oif(rt->rt));
}

int16_t nic_route_get_iif( IProute_t rt )
{
    if( rt == 0L ) return 0xfff;
    
    if( rt->rt == 0L )
	return rt->iif;
    
    char *iifn = rtnl_route_get_iif(rt->rt);

    if( iifn == 0L )
	return -1;
    
    struct rtnl_link *link = rtnl_link_get_by_name( rt->nh->lc,  iifn );
    
    rt->iif = rtnl_link_get_ifindex(link);
    
    rtnl_link_put(link);

    return rt->iif;
}

uint8_t nic_route_get_scope(IProute_t rt)
{
    if( rt == 0L ) return 0;
    
    if( rt->rt == 0L )
	return rt->scope;
    
    return (rt->scope = rtnl_route_get_scope(rt->rt));    
}

uint32_t nic_route_get_flags( IProute_t rt)
{    
    if( rt == 0L ) return 0;
    
    if( rt->rt == 0L )
	return rt->flags;
    
    return (rt->flags = rtnl_route_get_flags(rt->rt));
}

uint32_t nic_route_get_priority( IProute_t rt)
{    
    if( rt == 0L ) return 0;
    
    if( rt->rt == 0L )
	return rt->priority;
    
    return (rt->flags = rtnl_route_get_prio(rt->rt));
}

IPaddr_t nic_route_get_dst( IProute_t rt )
{
    if( rt == 0L )
	return 0;
    if( rt->dst_sa.sa_family )
	return 
	    nic_sa_addr(rt->nh, &(rt->dst_sa),
			(rt->dst_sa.sa_family == AF_INET) 
			? sizeof(struct in_addr)
			: sizeof(struct in6_addr)
		       );
    
    if( rt->rt == 0L )
	return 0L;

    struct nl_addr *nla = rtnl_route_get_dst( rt->rt );

    if( nla == 0 )
	return 0L;

    rt->dst_len = rtnl_route_get_dst_len( rt->rt );

    rt->dst_sa.sa_family = nl_addr_get_family( nla );

    if(  (rt->dst_sa.sa_family != AF_INET) 
       &&(rt->dst_sa.sa_family != AF_INET6)
      )
    {
	rt->dst_sa.sa_family = 0;
	return(0L);
    }
    
    uint8_t sa_len = nl_addr_get_len( nla );
    
    if( sa_len > sizeof(struct in6_addr) )
    {
	rt->dst_sa.sa_family = 0;
	return(0L);
    }
    
    uint8_t *ba = nl_addr_get_binary_addr( nla );

    if( ba == 0L )
	return 0L;

    if( rt->dst_sa.sa_family == AF_INET )
	memcpy(&(((struct sockaddr_in*)&(rt->dst_sa))->sin_addr), ba, sa_len);
    else
	memcpy(&(((struct sockaddr_in6*)&(rt->dst_sa))->sin6_addr), ba, sa_len);

    return nic_sa_addr(rt->nh, &(rt->dst_sa), sa_len);	
}

uint8_t nic_route_get_dst_len( IProute_t rt )
{
    if( rt == 0L )
	return 0;
    
    if( rt->dst_len )
	return rt->dst_len;

    if( rt->rt == 0 )
	return 0;

    return (rt->dst_len = rtnl_route_get_dst_len(rt->rt));    
}

IPaddr_t nic_route_get_gateway( IProute_t rt )
{
    if( rt == 0L )
	return 0;
    if( rt->gw_sa.sa_family )
	return 
	    nic_sa_addr(rt->nh, &(rt->gw_sa),
			(rt->gw_sa.sa_family == AF_INET) 
			? sizeof(struct in_addr)
			: sizeof(struct in6_addr)
		       );
    
    if( rt->rt == 0L )
	return 0L;

    struct nl_addr *nla = rtnl_route_get_gateway( rt->rt );

    if( nla == 0 )
	return 0L;

    rt->gw_sa.sa_family = nl_addr_get_family( nla );

    if(  (rt->gw_sa.sa_family != AF_INET) 
       &&(rt->gw_sa.sa_family != AF_INET6)
      )
    {
	rt->gw_sa.sa_family = 0;
	return(0L);
    }
    
    uint8_t sa_len = nl_addr_get_len( nla );
    
    if( sa_len > sizeof(struct in6_addr) )
    {
	rt->gw_sa.sa_family = 0;
	return(0L);
    }
    
    uint8_t *ba = nl_addr_get_binary_addr( nla );

    if( ba == 0L )
	return 0L;
    
    if( rt->gw_sa.sa_family == AF_INET )
	memcpy(&(((struct sockaddr_in*)&(rt->gw_sa))->sin_addr), ba, sa_len);
    else
	memcpy(&(((struct sockaddr_in6*)&(rt->gw_sa))->sin6_addr), ba, sa_len);

    return nic_sa_addr(rt->nh, &(rt->gw_sa), sa_len);	
}


IPaddr_t nic_route_get_src( IProute_t rt )
{
    if( rt->src_sa.sa_family )
	return 
	    nic_sa_addr(rt->nh, &(rt->src_sa),
			(rt->src_sa.sa_family == AF_INET) 
			? sizeof(struct in_addr)
			: sizeof(struct in6_addr)
		       );
    
    if( rt->rt == 0L )
	return 0L;

    struct nl_addr *nla = rtnl_route_get_src( rt->rt );

    if( nla == 0 )
	return 0L;

    rt->src_len = rtnl_route_get_src_len( rt->rt );

    rt->src_sa.sa_family = nl_addr_get_family( nla );

    if(  (rt->src_sa.sa_family != AF_INET) 
       &&(rt->src_sa.sa_family != AF_INET6)
      )
    {
	rt->src_sa.sa_family = 0;
	return(0L);
    }
    
    uint8_t sa_len = nl_addr_get_len( nla );
    
    if( sa_len > sizeof(struct in6_addr) )
    {
	rt->src_sa.sa_family = 0;
	return(0L);
    }
    
    uint8_t *ba = nl_addr_get_binary_addr( nla );

    if( ba == 0L )
	return 0L;
    
    if( rt->src_sa.sa_family == AF_INET )
	memcpy(&(((struct sockaddr_in*)&(rt->src_sa))->sin_addr), ba, sa_len);
    else
	memcpy(&(((struct sockaddr_in6*)&(rt->src_sa))->sin6_addr), ba, sa_len);

    return nic_sa_addr(rt->nh, &(rt->src_sa), sa_len);	
}

uint8_t nic_route_get_src_len( IProute_t rt )
{
    if( rt == 0L )
	return 0;
    
    if( rt->src_len )
	return rt->src_len;

    if( rt->rt == 0 )
	return 0;

    return (rt->src_len = rtnl_route_get_src_len(rt->rt));    
}

IProute_t nic_route_new
(   NLH_t nh,
    int16_t oif,
    struct sockaddr *dst_sa,
    uint8_t  dst_len, 
    struct sockaddr *gw_sa,
    int8_t scope,
    int32_t priority,
    int32_t table,
    int16_t iif,
    struct sockaddr *src_sa,
    uint8_t  src_len
)
{
    uint8_t default_route = 0;
    if( ( dst_sa && (dst_sa->sa_family != AF_INET) && (dst_sa->sa_family != AF_INET6))
      ||( gw_sa && (gw_sa->sa_family != AF_INET) && (gw_sa->sa_family != AF_INET6))
      ||( src_sa && (src_sa->sa_family != AF_INET) && (src_sa->sa_family != AF_INET6))
      )
	return 0L;
	
    IProute_t rt = nic_calloc(1, sizeof(struct nic_route_s));

    if(rt == 0L)
	return rt;
    
    rt->nh = nh;
    if((rt->rt = rtnl_route_alloc())==0L)
    {
	nic_fre(rt);
	return 0L;
    }

    if( oif != -1 )
	rtnl_route_set_oif(rt->rt, rt->oif = oif);
    else
	rt->oif = -1;
   
    if ( dst_sa == 0L )
    { /* default route */
	dst_len = 0;
	rt->dst_len = 0;
	rt->family = AF_INET;
	rt->dst_sa.sa_family = AF_INET;
	table = RT_TABLE_MAIN;
	default_route = 1;
    }else
    {
	if ( dst_sa -> sa_family == AF_INET )
	{
	    rt->family = AF_INET;
	    ((struct sockaddr_in*)(&(rt->dst_sa)))->sin_family = AF_INET;
	    memcpy(&(((struct sockaddr_in*)&(rt->dst_sa))->sin_addr),
		   &(((struct sockaddr_in*)dst_sa)->sin_addr),
		   sizeof(struct in_addr)
		  );
	}else 
   	if ( dst_sa -> sa_family == AF_INET6 )
	{
	    rt->family = AF_INET6;
	    ((struct sockaddr_in6*)(&(rt->dst_sa)))->sin6_family = AF_INET6;
	    memcpy(&(((struct sockaddr_in6*)&(rt->dst_sa))->sin6_addr),
		   &(((struct sockaddr_in6*)dst_sa)->sin6_addr),
		   sizeof(struct in6_addr)
		  );
	}else
	    dst_len = 0;
    }
    
    if ( ( dst_len || default_route ) && (rt->dst_sa.sa_family) )
    {
	struct nl_addr *nla = nl_addr_alloc(sizeof(struct sockaddr_in6));
	
	nl_addr_set_family(nla, rt->dst_sa.sa_family );

	if( rt->dst_sa.sa_family == AF_INET )
	    nl_addr_set_binary_addr
	    (   nla,
		&(((struct sockaddr_in*)&(rt->dst_sa))->sin_addr),
		sizeof(struct in_addr)
	    );
	else
	    nl_addr_set_binary_addr
	    (   nla,
		&(((struct sockaddr_in6*)&(rt->dst_sa))->sin6_addr),
		sizeof(struct in6_addr)
	    );
	
	rt->dst_len = dst_len;

	nl_addr_set_prefixlen(nla, dst_len);
	
	rtnl_route_set_dst(rt->rt, nla);
    }

    if( gw_sa  )
    {	
	struct nl_addr *nla = nl_addr_alloc(sizeof(struct sockaddr_in6));
	
	nl_addr_set_family
        (   nla, 
	    rt->gw_sa.sa_family = gw_sa->sa_family 
	);

	if( gw_sa->sa_family == AF_INET )
	{
	    memcpy
	    (   &(((struct sockaddr_in*)&(rt->gw_sa))->sin_addr),
		&(((struct sockaddr_in*)gw_sa)->sin_addr),
		sizeof(struct in_addr)
	    );
	    nl_addr_set_binary_addr
	    (   nla,
		&(((struct sockaddr_in*)gw_sa)->sin_addr),
		sizeof(struct in_addr)
	    );
	}else
	{
	    memcpy
	    (   &(((struct sockaddr_in6*)&(rt->gw_sa))->sin6_addr),
		&(((struct sockaddr_in6*)gw_sa)->sin6_addr),
		sizeof(struct in6_addr)
	    );
	    nl_addr_set_binary_addr
	    (   nla,
		&(((struct sockaddr_in6*)gw_sa)->sin6_addr),
		sizeof(struct in6_addr)
	    );	
	}
	rtnl_route_set_gateway(rt->rt, nla);	
    }

    if( scope == -1 )
	rt->scope = RT_SCOPE_UNIVERSE;
    
    rtnl_route_set_scope(rt->rt, rt->scope);
	
    if( iif != -1 )
    {
	struct rtnl_link *link = rtnl_link_get(nh->lc, iif);
	if( link )
	    rtnl_route_set_iif(rt->rt, rtnl_link_get_name(link));
	rtnl_link_put(link);
    }else
	rt->iif = -1;

    if( src_sa && src_len )
    {	
	struct nl_addr *nla = nl_addr_alloc(sizeof(struct sockaddr_in6));
	
	nl_addr_set_family
        (   nla,
	    rt->src_sa.sa_family = src_sa->sa_family 
	);

	if( rt->src_sa.sa_family == AF_INET )
	{
	    memcpy( &(((struct sockaddr_in*)&(rt->src_sa))->sin_addr),
		    &(((struct sockaddr_in*)src_sa)->sin_addr),
		    sizeof(struct in_addr)
		  );
	    nl_addr_set_binary_addr
	    (   nla,
		&(((struct sockaddr_in*)&(rt->src_sa))->sin_addr),
		sizeof(struct in_addr)
	    );
	}else
	{
	    memcpy( &(((struct sockaddr_in6*)&(rt->src_sa))->sin6_addr),
		    &(((struct sockaddr_in6*)src_sa)->sin6_addr),
		    sizeof(struct in6_addr)
		  );
	    nl_addr_set_binary_addr
	    (   nla,
		&(((struct sockaddr_in6*)&(rt->src_sa))->sin6_addr),
		sizeof(struct in6_addr)
	    );
	}

	rt->src_len = src_len;

	nl_addr_set_prefixlen(nla, src_len);

	rtnl_route_set_src(rt->rt, nla);	
    }

    if( priority != -1 )
	rtnl_route_set_prio(rt->rt, rt->priority = priority);
    
    rtnl_route_set_table
    (   rt->rt, 
	rt->table =
	(( table == -1 )
	 ? rtnl_route_str2table("local")
	 : table
	)
    );
    
    return rt;
}

void nic_route_free( IProute_t rt )
{
    if( rt == 0 ) return;
    
    if( rt->rt )
	rtnl_route_put(rt->rt);
    
    nic_fre(rt);
}

static NIC_Res_t nic_route_msg( IProute_t rt, char *fn, char *tag, uint32_t msg_type, uint32_t msg_flags)
{
    /* libnl as yet provides no support for route creation / deletion / modification.
     * We'll have to roll our own:
     */
    struct nl_msg *msg = nlmsg_build_simple(msg_type, msg_flags) ;
    if( msg == 0L )
    {
	if (rt->nh->eh)
	    rt->nh->eh(rt->nh,NIC_ERR,"%s: failed to %s route - out of memory.", tag, fn);
	return NIC_FAIL;
    }
    struct rtmsg rtm ;
    memset(&rtm, '\0', sizeof(rtm));
    
    rtm.rtm_type     = RTN_UNICAST;
    rtm.rtm_family   = rt->family;
    rtm.rtm_dst_len  = rt->dst_len;
    rtm.rtm_src_len  = rt->src_len;
    rtm.rtm_table    = rt->table;
    rtm.rtm_scope    = rt->scope;
    rtm.rtm_protocol = RTPROT_BOOT;
    
    if ( nlmsg_append(msg,&rtm,sizeof(rtm),1) < 0 )
	goto nla_put_failure;

    if ( rt->oif != -1 )
    {
	uint32_t oif = rt->oif;
	NLA_PUT_U32(msg, RTA_OIF, oif);
    }
    
    if ( rt->dst_len )
    {
	struct nl_addr *nla = rtnl_route_get_dst(rt->rt);
	
	NLA_PUT_ADDR(msg, RTA_DST, nla);
    }

    if ( rt->gw_sa.sa_family )
    {
	struct nl_addr *nla = rtnl_route_get_gateway(rt->rt);

	NLA_PUT_ADDR(msg, RTA_GATEWAY, nla);
    }

    if ( rt->priority != -1 )
    {
	NLA_PUT_U32( msg, RTA_PRIORITY, rt->priority);
    }
        
    if ( rt->src_len )
    {
	struct nl_addr *nla = rtnl_route_get_src(rt->rt);
	
	NLA_PUT_ADDR(msg, RTA_SRC, nla);
    }

    if ( rt->iif != -1 )
    {
	struct rtnl_link *link = rtnl_link_get(rt->nh->lc, rt->iif);
	char *name=rtnl_link_get_name(link);	
	NLA_PUT_STRING(msg, RTA_IIF, name);
	rtnl_link_put(link);
    }

    if( nl_send_auto_complete(rt->nh->nl, nlmsg_hdr(msg)) < 0 )
    {
	if (rt->nh->eh)
	    rt->nh->eh(rt->nh,NIC_ERR,"%s: send failed - %d %s", tag,
		       nl_get_errno(), nl_geterror()
		      );
	return NIC_FAIL;
    }

    nlmsg_free(msg);

    if( nl_wait_for_ack(rt->nh->nl) < 0 )
    {
	if (rt->nh->eh)
	    rt->nh->eh(rt->nh,NIC_ERR,"%s: failed - %d: %s", tag,
		       nl_get_errno(), nl_geterror()
		      );
	return NIC_FAIL;	
    }
    
    return NIC_OK;

nla_put_failure:

    if (rt->nh->eh)
	rt->nh->eh(rt->nh,NIC_ERR,"%s: failed to %s route - out of memory.", tag, fn);
    nlmsg_free(msg);
    return NIC_FAIL;	
}

NIC_Res_t nic_add_route( IProute_t rt )
{ 
    if ((rt == 0L) || (rt->rt == 0L)) return NIC_FAIL;

    return nic_route_msg( rt, "nic_add_route", "add", RTM_NEWROUTE, NLM_F_CREATE | NLM_F_REPLACE);
}

NIC_Res_t nic_update_route( IProute_t rt )
{ 
    if ((rt == 0L) || (rt->rt == 0L)) return NIC_FAIL;

    return nic_route_msg( rt, "nic_add_route", "add", RTM_NEWROUTE, NLM_F_REPLACE);
}

NIC_Res_t nic_remove_route( IProute_t rt )
{ 
    if ((rt == 0L) || (rt->rt == 0L)) return NIC_FAIL;

    return nic_route_msg( rt, "nic_add_route", "add", RTM_DELROUTE, 0);
}

IProute_list_t *nic_route_list_new( IProute_t rt1, ... )
{      
    va_list va;
    va_start(va, rt1);
    
    IProute_list_t *head = nic_calloc(1,sizeof(IProute_list_t));
    IProute_list_node_t *n;
    IProute_t rt;

    if( head == 0 )
    {
	va_end(va);
	if( rt1->nh->eh )
	    rt1->nh->eh ( rt1->nh, NIC_ERR, "out of memory" );
	return 0;
    }
    rt=rt1;
    STAILQ_INIT(head);
    do
    {
	n = nic_calloc(1,sizeof(IProute_list_node_t));
	if ( n == 0L )
	{
	    va_end(va);
	    nic_fre(head);
	    if( rt->nh->eh )
		rt->nh->eh ( rt->nh, NIC_ERR, "out of memory" );
	    return 0L;
	}
	n->route = rt;
	STAILQ_INSERT_TAIL( head, n, link );
    }while( (rt = va_arg(va, IProute_t)) != 0) ;
    va_end(va);
    return(head);
}

void nic_route_list_free( IProute_list_t *head )
{
    IProute_list_node_t *n, *a;
    
    if ( !STAILQ_EMPTY(head) )
    {
	n = STAILQ_FIRST( head );
	do
	{
	    a = STAILQ_NEXT(n, link);
	    nic_route_free(n->route);
	    nic_fre(n);
	    n = a;
	}while (n);
    }
}

NIC_Res_t nic_add_routes( IProute_list_t *head )
{
    IProute_list_node_t *n;
    NIC_Res_t r;
    STAILQ_FOREACH(n,head,link)
    {
	if((r=nic_add_route(n->route)) == NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

NIC_Res_t nic_remove_routes( IProute_list_t *head )
{
    IProute_list_node_t *n;
    NIC_Res_t r;
    STAILQ_FOREACH(n,head,link)
    {
	if((r=nic_remove_route(n->route))==NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

NIC_Res_t nic_update_routes( IProute_list_t *head )
{
    IProute_list_node_t *n;
    NIC_Res_t r;
    STAILQ_FOREACH(n,head,link)
    {
	if((r=nic_update_route(n->route))==NIC_FAIL)
	    return r;
    }
    return NIC_OK;
}

NIC_Res_t 
nic_configure_resolver
(
    NLH_t           nh,
    IPaddr_list_t  *dns_servers,
    char           *search_list
)
{
    if( (nh == 0) || 
	(  (dns_servers == 0L) 
        && ((search_list==0L) || (*search_list =='\0'))
	)
      ) return NIC_FAIL;    	
    char tmpName[] = "/etc/XXXXXX";
    int fd = mkstemp (&(tmpName[0]));
    FILE *fp;
    IPaddr_list_node_t *n;
    char buf[64];
    if( fd == -1 )
	return NIC_FAIL;
    if( (fp = fdopen(fd,"w")) == 0L )
	return NIC_FAIL;
    fprintf(fp,
	    "; resolv.conf generated by libdhcp\n"
	    "search %s\n",
	    search_list
	   );
    STAILQ_FOREACH(n,dns_servers,link)
    {
	if( n->addr )
	    fprintf(fp,"nameserver %s\n",nic_addr_to_text(n->addr,buf,64));
    }
    fflush(fp);
    fclose(fp);
    if( rename("/etc/resolv.conf", "/etc/resolv.conf.predhclient") != 0 )
    {
	unlink(tmpName);
	return NIC_FAIL;
    }
    if( rename(tmpName,"/etc/resolv.conf") != 0 )
    {
	unlink(tmpName);
	return NIC_FAIL;
    }
    return NIC_OK;
}

extern
NIC_Res_t 
nic_set_hostname
( NLH_t nh,
  char *hostname
)
{
    return sethostname(hostname, strlen(hostname)) == 0 ? NIC_OK : NIC_FAIL;
}

NIC_Res_t 
nic_configure
(
    NLH_t          nh,
    NIC_t          nic,
    IPaddr_list_t  *addresses,
    IProute_list_t *routes,
    IPaddr_list_t  *dns_servers,
    char           *search_list,
    char           *host_name
)
{
    if ( addresses && ( nic_add_addresses( nic, addresses ) == NIC_FAIL ))
    {
	if (nh->eh)
	    nh->eh(nh, NIC_ERR, "nic_configure: failed to add addresses.");
	return NIC_FAIL;
    }
     
    if ( routes && ( nic_add_routes( routes ) == NIC_FAIL ))
    {
	if (nh->eh)
	    nh->eh(nh, NIC_ERR, "nic_configure: failed to add routes.");
	return NIC_FAIL;
    }

    if (   ( search_list || dns_servers ) 
        && ( nic_configure_resolver( nh, dns_servers, search_list ) == NIC_FAIL )
       )
    {
	if (nh->eh)
	    nh->eh(nh, NIC_ERR, "nic_configure: failed to configure resolver.");
	return NIC_FAIL;
    }
    return NIC_OK;
}
