/* nic.h
 *
 *  Network Interface Configurator for libdhcp .
 * 
 *  Provides a simplified interface to the 'libnl' Netlink Library, and
 *  functions which configure an interface from the dhcp4_lease and 
 *  dhcp6_lease objects provided by libdhcp. This header does not
 *  require the libnl headers, but each object provides a function
 *  for obtaining the libnl opaque pointer with which any libnl
 *  function can be called should users choose to include the
 *  libnl headers. 
 *
 *  Copyright(C) Jason Vas Dias <jvdias@redhat.com> Red Hat Inc. May 2006
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation at 
 *           http://www.fsf.org/licensing/licenses/gpl.txt
 *  and included in this software distribution as the "LICENSE" file.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 */
#ifndef LIBDHCP_NIC_H
#define LIBDHCP_NIC_H

#include <sys/types.h>
#include <sys/queue.h>
#include <netinet/in.h>
#include <net/if.h>
#include <stdint.h>
#include <stdarg.h>

/*
 * Network Interface Configuration (NIC) Link opaque type:
 */
struct nic_s;
typedef struct nic_s  *NIC_t;

/*
 *  Interface Address and Subnet opaque type :
 */
struct  nic_ip_address_s;
typedef struct nic_ip_address_s *IPaddr_t;

typedef
struct nic_ip_address_list_node_s
{
    IPaddr_t addr;
    STAILQ_ENTRY( nic_ip_address_list_node_s ) link;
} IPaddr_list_node_t;

typedef 
STAILQ_HEAD(nic_ip_address_list_s,nic_ip_address_list_node_s)
IPaddr_list_t;

/*  
 *  Interface Route opaque type:
 */
struct nic_route_s;
typedef struct nic_route_s *IProute_t;

typedef
struct nic_ip_route_node_s
{
    IProute_t route;
    STAILQ_ENTRY( nic_ip_route_node_s ) link;
} IProute_list_node_t;

typedef
STAILQ_HEAD(nic_ip_route_list_s,nic_ip_route_node_s)
IProute_list_t;

/* Opaque NIC Library Handle - users must first obtain one to create 
 * any NIC object. Also records error handler and log level:
 */
typedef struct nlh_s *NLH_t; 

typedef 
enum nic_error_level_e
{ /* syslog priority */
    NIC_FATAL,    /* cannot continue - unused!   */
    NIC_ERR=3,    /* operation failed            */
    NIC_WARN=4,   /* not all OK with operation   */
    NIC_DEBUG=7 
} NIC_Error_Level_t;

typedef
enum nic_result_e
{   /* < -1 : -1 *netlink error code */
    NIC_FAIL = -1,
    NIC_OK   =  0,
    NIC_SUCCESS=0
    /* > 0: OK status  */
} NIC_Res_t;


typedef void (*NIC_Error_Handler_t)( NLH_t, NIC_Error_Level_t, char*, ...); 

extern
NLH_t nic_open( NIC_Error_Handler_t );

extern
void nic_close(NLH_t*); 

/* default error handlers: */

extern
void nic_set_logger(NLH_t, NIC_Error_Handler_t /* can be NULL: no logging - default */ );

typedef int (*NIC_VA_Error_Handler_t)(void*, int priority, char *fmt, va_list va);

extern
void nic_set_va_logger(NLH_t, NIC_VA_Error_Handler_t handler, void *handler_arg );

extern
void nic_sys_logger( NLH_t, NIC_Error_Level_t, char *fmt, ... );

extern
void nic_stderr_logger( NLH_t, NIC_Error_Level_t, char *fmt, ... );

extern
NIC_Res_t nic_set_loglevel( NLH_t, NIC_Error_Level_t );

/*
 *  Interface NIC (Link / IF) functions:
 */
extern
NIC_t nic_by_name(NLH_t, char * );   /* constructor */

extern
NIC_t nic_by_index(NLH_t, uint16_t );/* constructor */

typedef void (*NIC_handler_t)(NLH_t nh, NIC_t, void*);

extern  void nic_foreach(NLH_t nh, NIC_handler_t handler,  void *arg);
/* Calls handler with nic of each interface currently configured in
 * the kernel and user void* arg arguments. 
 * NOTE: the callback MUST NOT free(nic) ! 
 */

extern
char  *nic_get_name( NIC_t );

extern
uint16_t nic_get_index( NIC_t );

extern
IPaddr_t nic_get_link_addr(NIC_t);
/* get link layer / ethernet address
 */

extern
IPaddr_t nic_get_link_broadcast(NIC_t);
/* get link layer broadcast address
 */

extern
uint32_t nic_get_flags( NIC_t );

extern
NIC_Res_t nic_set_flags( NIC_t, uint32_t  );

extern
uint32_t nic_get_mtu( NIC_t );

extern
NIC_Res_t nic_set_mtu( NIC_t, uint32_t  );

struct rtnl_link *nic_rtnl_link_get(NIC_t);
void nic_rtnl_link_put(struct rtnl_link *);
/* Every other Link attribute can be set with 
 * the libnl rtnl_link module - include the
 * libnl headers to use, and call rtnl_link_put
 * when finished with the pointer.
 */

extern
void  nic_free( NIC_t );       /* destructor */


/*
 *  NIC IP Address functions:
 */

typedef void (*IPaddr_Handler_t) (NLH_t nh, IPaddr_t, void *);

void nic_addr_foreach( NLH_t, IPaddr_Handler_t, void * );
/* calls handler for every address currently configured */

void nic_addr_nic_foreach( NLH_t, NIC_t, IPaddr_Handler_t, void * );
/* calls handler for every address currently configured on NIC */

extern
IPaddr_t nic_sa_addr(NLH_t, struct sockaddr *sa, socklen_t sz );

extern
struct sockaddr *nic_addr_sa(  IPaddr_t, socklen_t *sz );

uint8_t nic_addr_get_family( IPaddr_t );

uint8_t nic_addr_get_prefix( IPaddr_t );

void nic_addr_set_prefix( IPaddr_t, uint8_t );

extern
IPaddr_t nic_addr_from_text(NLH_t, char* ); 
/* inic_ntop + (ipv4|ipv6) + [/pfx] prefix length */

extern
char    *nic_addr_to_text(IPaddr_t, char *, size_t);

extern
IPaddr_t nic_ipv4_addr( NLH_t, struct in_addr );

extern
IPaddr_t nic_ipv6_addr( NLH_t, struct in6_addr );

extern
IPaddr_t nic_ipv4_broadcast_addr( IPaddr_t );

extern
NIC_Res_t nic_add_address( NIC_t, IPaddr_t );

extern
NIC_Res_t nic_remove_address( NIC_t, IPaddr_t );

extern
NIC_Res_t nic_addr_get_broadcast( IPaddr_t, struct sockaddr*, size_t sz );

extern
void nic_addr_set_broadcast( IPaddr_t, IPaddr_t );

extern
int8_t nic_addr_get_scope(IPaddr_t);

extern
void nic_addr_set_scope(IPaddr_t, int8_t);

extern
IPaddr_list_t *nic_address_list_new( IPaddr_t, ... );
/* All args in list must be IPaddr_t; last arg MUST be 0
 */
extern
void nic_address_list_free( IPaddr_list_t *);

extern
NIC_Res_t nic_add_addresses( NIC_t, IPaddr_list_t* );

extern
NIC_Res_t nic_remove_addresses( NIC_t, IPaddr_list_t* );

struct rtnl_addr *nic_rtnl_addr_get(IPaddr_t);
void nic_rtnl_addr_put(struct rtnl_addr *);
struct nl_addr *nic_nl_addr_get(IPaddr_t);
void nic_nl_addr_put(struct nl_addr *);
/* Every other IFA attribute can be set with 
 * the libnl rtnl_addr module - include the
 * libnl headers to use, and call rtnl_addr_put
 * when finished with the pointer.
 */

void nic_addr_free( IPaddr_t );

/*
 *  NIC IP Route functions: 
 */

typedef void (*IProute_handler_t) (NLH_t nh, IProute_t, void *);

extern
void nic_route_foreach( NLH_t, IProute_handler_t, void * );
/* calls handler for every route currently configured */

extern
int32_t nic_route_get_table(IProute_t);

extern
char *nic_route_get_table_name( int32_t, char *buf, int len );

extern
int32_t nic_route_get_table_number( char* );

extern
uint8_t nic_route_get_family( IProute_t );

extern
int16_t nic_route_get_oif( IProute_t );
/* oif = output interface */

extern
int16_t nic_route_get_iif( IProute_t );

extern
uint8_t nic_route_get_scope(IProute_t);

extern
uint32_t nic_route_get_flags( IProute_t );

extern
uint32_t nic_route_set_priority( IProute_t );

extern 
IPaddr_t nic_route_get_dst( IProute_t );

extern
uint8_t nic_route_get_dst_len( IProute_t );

extern 
IPaddr_t nic_route_get_src( IProute_t );

extern
uint8_t nic_route_get_src_len( IProute_t );

extern 
IPaddr_t nic_route_get_gateway( IProute_t );

extern
IProute_t nic_route_new
(   NLH_t,
    int16_t oif,          /* output interface: -1 means none */
    struct sockaddr *dst, /* 0L here means DEFAULT ROUTE */ 
#define      NIC_ROUTE_DEFAULT_ADDR ((struct sockaddr*)0L)
    uint8_t  dst_len,     /* dst_len ignored for default route */
    struct sockaddr *gw,  /* 0L here means no gateway */
    int8_t scope,         /* -1=UNIVERSE,LINK=253,SITE=200,HOST=254 */
    int32_t priority,     /* metric - -1 means no priority */
    int32_t table,        /* table: -1 selects default: "local" */
    int16_t iif,          /* input interface: -1 means none */
    struct sockaddr *src, /* 0L here means no src route */
    uint8_t  src_len      /* 0 here means no src route */
#define      NIC_ROUTE_NO_IF (-1)
);

extern
void nic_route_free( IProute_t );

extern
NIC_Res_t nic_add_route( IProute_t );

extern
NIC_Res_t nic_update_route( IProute_t );

extern
NIC_Res_t nic_remove_route( IProute_t );

extern
IProute_list_t *nic_route_list_new( IProute_t, ... );
/* All args in list must be IProute_t; last arg MUST be 0
 */
void nic_route_list_free( IProute_list_t* );

extern
NIC_Res_t nic_add_routes( IProute_list_t* );

extern
NIC_Res_t nic_remove_routes( IProute_list_t* );

extern
NIC_Res_t nic_update_routes( IProute_list_t* );

struct rtnl_route *nic_rtnl_route_get(IProute_t);
void nic_rtnl_route_put(struct rtnl_route *);
/* Every other RTA attribute can be set with 
 * the libnl rtnl_route module - include the
 * libnl headers to use, and call rtnl_route_put
 * when finished with the pointer.
 */
extern void nic_free_route( IProute_t );

/*
 * Global Configuration Operations:
 */

extern
NIC_Res_t 
nic_configure_resolver
(
    NLH_t          nh,
    IPaddr_list_t  *dns_servers,
    char           *search_list
);

extern
NIC_Res_t 
nic_set_hostname
( NLH_t nh,
  char *hostname
);

extern
NIC_Res_t 
nic_configure
(
    NLH_t          nh,
    NIC_t          nic,
    IPaddr_list_t  *addresses,
    IProute_list_t *routes,
    IPaddr_list_t  *dns_servers,
    char           *search_list,
    char           *host_name
);



#endif
